from ploomber import load_env

import warnings
import smtplib
from email.message import EmailMessage


@load_env
def send_email(env, subject, body):
    """Send an email notification
    """
    msg = EmailMessage()
    msg.set_content(body)

    try:
        sender = env.notifications.email.sender
    except KeyError:
        sender = None

    try:
        to = env.notifications.email.to
    except KeyError:
        to = None

    if sender is not None and to is not None:
        msg['Subject'] = subject
        msg['From'] = sender
        msg['To'] = env.notifications.email.to

        s = smtplib.SMTP('localhost')
        s.send_message(msg)
        s.quit()
    else:
        if sender is None:
            warnings.warn('Cannot send email, could not load sender')

        if to is None:
            warnings.warn('Cannot send email, could not load recipient')



import logging
import warnings
import abc
from datetime import datetime
from ploomber.util.util import callback_check


class AbstractMetadata(abc.ABC):
    """Abstract class to represent Product's metadata

    If product does not exist, initialize empty metadata, otherwise use
    product.fetch_metadata, and accept it after doing some validations
    """

    def __init__(self, product):
        # NOTE: cannot fetch initial metadata here, since product is not
        # rendered on Metadata initialization, we have to do lazy loading
        self._data = None
        self._product = product

        self._logger = logging.getLogger('{}.{}'.format(__name__,
                                                        type(self).__name__))

    @property
    @abc.abstractmethod
    def timestamp(self):
        """When the product was originally created
        """
        pass

    @property
    @abc.abstractmethod
    def stored_source_code(self):
        """Source code that generated the product
        """
        pass

    @abc.abstractmethod
    def update(self, source_code):
        """
        """
        pass

    @abc.abstractmethod
    def load(self):
        """Load metadata
        """
        pass

    def __eq__(self, other):
        return self.data == other

    def __getstate__(self):
        state = self.__dict__.copy()

        if '_logger' in state:
            del state['_logger']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger('{}.{}'.format(__name__,
                                                        type(self).__name__))


class Metadata(AbstractMetadata):
    """
    Internal class to standardize access to Product's metadata
    """
    # TODO: this should be _data, since it is an internal API
    @property
    def data(self):
        if self._data is None:
            self.load()
        return self._data

    @property
    def timestamp(self):
        return self.data.get('timestamp')

    @property
    def stored_source_code(self):
        return self.data.get('stored_source_code')

    def load(self):
        if not self._product.exists():
            metadata = dict(timestamp=None, stored_source_code=None)
        else:
            metadata_fetched = self._product.fetch_metadata()

            if metadata_fetched is None:
                self._logger.debug('fetch_metadata for product %s returned '
                                   'None',
                                   self._product)
                metadata = dict(timestamp=None, stored_source_code=None)
            else:
                # FIXME: we need to further validate this, need to check
                # that this is an instance of mapping, if yes, then
                # check keys [timestamp, stored_source_code], check
                # types and fill with None if any of the keys is missing
                metadata = metadata_fetched

        self._data = metadata

    def update(self, source_code):
        """
        """
        self.data['timestamp'] = datetime.now().timestamp()
        self.data['stored_source_code'] = source_code

        kwargs = callback_check(self._product.prepare_metadata,
                                available={'metadata': self.data,
                                           'product': self._product})

        data = self._product.prepare_metadata(**kwargs)

        self._product.save_metadata(data)

    def __getitem__(self, key):
        return self._data[key]


class MetadataCollection(AbstractMetadata):
    """Metadata class used for MetaProduct
    """

    def __init__(self, products):
        self._products = products

    @property
    def timestamp(self):
        """When the product was originally created
        """
        # TODO: refactor, all products should have the same metadata
        timestamps = [p.metadata.timestamp
                      for p in self._products
                      if p.metadata.timestamp is not None]
        if timestamps:
            return max(timestamps)
        else:
            return None

    @property
    def stored_source_code(self):
        """Source code that generated the product
        """
        stored_source_code = set([p.metadata.stored_source_code
                                  for p in self._products
                                  if p.metadata.stored_source_code
                                  is not None])
        if len(stored_source_code):
            warnings.warn('Stored source codes for products {} '
                          'are different, but they are part of the same '
                          'MetaProduct, returning stored_source_code as None'
                          .format(self._products))
            return None
        else:
            return list(stored_source_code)[0]

    def update(self, source_code):
        """
        """
        for p in self._products:
            p.metadata.update(source_code)

    def load(self):
        for p in self._products:
            p.metadata.load()

    # TODO: add getitem


class MetadataAlwaysUpToDate(AbstractMetadata):
    """
    Metadata for Link tasks (always up-to-date)
    """

    def __init__(self):
        pass

    @property
    def timestamp(self):
        return 0

    @property
    def stored_source_code(self):
        return None

    def load(self):
        pass

    def update(self, source_code):
        pass

    # TODO: add getitem



"""
Notes to developers

meta:
  Settings that cannot be clearly mapped to the OOP Python interface, we
  don't call it config because there is a DAGConfig object in the Python
  API and this might cause confusion


All other sections should represent valid DAG properties.
"""
import os
import sys
import yaml
import logging
from pathlib import Path
from collections.abc import MutableMapping

from ploomber import products
from ploomber import DAG, tasks
from ploomber.util.util import load_dotted_path, find_file_recursively
from ploomber.static_analysis import project
from ploomber.spec.TaskSpec import TaskSpec
from ploomber.spec import validate
from ploomber.dag.DAGConfiguration import DAGConfiguration
from ploomber.exceptions import DAGSpecInitializationError

logger = logging.getLogger(__name__)


class DAGSpec(MutableMapping):
    """
    A DAG spec is a dictionary with certain structure that can be converted
    to a DAG using DAGSpec.to_dag().

    There are two cases: the simplest one is just a dictionary with a
    "location" key with the factory to call, the other explicitly describes
    the DAG structure as a dictionary.

    Notes
    -----
    Relative paths are heavily used for pointing to source files and products,
    before the introduction of the Spec API, this was unambiguous: all paths
    were relative to the current working directory, you only had to make sure
    to start the session at the right place. Initially, the Spec API didn't
    change a thing since the assumption was that one would convert a spec to
    a DAG in a folder containing a pipeline.yaml file and paths in such file
    would be relative to the file itself, no ambiguity. However, when we
    introduced the Jupyter custom contents manager things changed because the
    DAG can be initialized from directories where there isn't a pipeline.yaml
    but we have to recursively look for one. Furthermore, if extract_upstream
    or extract_product are True and the source files are not in the same folder
    as pipeline.yaml an ambiguity arises: are paths inside a source file
    relative to the source file or pipeline.yaml? To make this consistent,
    all relative paths (whether in a pipeline.yaml or in a source file) are
    relative to the pipeline.yaml folder but this can be changed with the
    product_relative_to_source option. Furthermore, when usign the custom
    contents manager, we convert relative paths to absolute in special cases
    to make sure they work. This is why we need to keep the pipeline.yaml
    location when initializing a DAG from a spec.
    """
    def __init__(self, data):
        # only set when initialized from DAGSpec.from_file
        self._parent_path = None

        if isinstance(data, list):
            data = {'tasks': data}

        load_from_factory = self._validate_top_level_keys(data)
        self.data = data

        if not load_from_factory:
            self.data['tasks'] = [
                normalize_task(task) for task in self.data['tasks']
            ]
            self._validate_meta()

            self.data['tasks'] = [
                TaskSpec(t, self.data['meta']) for t in self.data['tasks']
            ]

    def _validate_top_level_keys(self, spec):
        load_from_factory = False

        if 'location' in spec:
            if len(spec) > 1:
                raise KeyError('If specifying dag through a "location" key '
                               'it must be the unique key in the spec')
            else:
                load_from_factory = True
        else:
            valid = {'meta', 'config', 'clients', 'tasks'}
            validate.keys(valid, spec.keys(), name='dag spec')

        return load_from_factory

    def _validate_meta(self):
        if 'meta' not in self.data:
            self.data['meta'] = {}

        self.default_meta(self.data['meta'])

    @classmethod
    def default_meta(cls, meta=None):
        """Fill missing values in a meta dictionary
        """
        if meta is None:
            meta = {}

        valid = {
            'extract_upstream', 'extract_product', 'product_default_class',
            'product_relative_to_source', 'jupyter_hot_reload'
        }
        validate.keys(valid, meta, name='dag spec')

        if 'extract_upstream' not in meta:
            meta['extract_upstream'] = True

        if 'extract_product' not in meta:
            meta['extract_product'] = True

        if 'product_relative_to_source' not in meta:
            meta['product_relative_to_source'] = False

        if 'jupyter_hot_reload' not in meta:
            meta['jupyter_hot_reload'] = False

        defaults = {
            'SQLDump': 'File',
            'NotebookRunner': 'File',
            'SQLScript': 'SQLRelation'
        }

        if 'product_default_class' not in meta:
            meta['product_default_class'] = defaults
        else:
            for class_, prod in defaults.items():
                if class_ not in meta['product_default_class']:
                    meta['product_default_class'][class_] = prod

        return meta

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __iter__(self):
        for key in self.data:
            yield key

    def __len__(self):
        return len(self.data)

    def to_dag(self):
        """Converts the DAG spec to a DAG object
        """
        cwd_old = os.getcwd()

        # when initializing DAGs from pipeline.yaml files, we have to ensure
        # that the parent folder is in sys.path for imports to work, this
        # happens most of the time but for some (unknown) reason, it doesn't
        # happen when initializing PloomberContentsManager.
        # pipeline.yaml paths are written relative to that file, for source
        # scripts to be located we temporarily change the current working
        # directory
        if self._parent_path is not None:
            sys.path.append(self._parent_path)
            os.chdir(self._parent_path)

        try:
            dag = self._to_dag()
        finally:
            if self._parent_path is not None:
                sys.path.remove(self._parent_path)
                os.chdir(cwd_old)
        return dag

    def _to_dag(self):
        # FIXME: validate that if there is location, there isn't anything else
        if 'location' in self:
            factory = load_dotted_path(self['location'])
            return factory()

        tasks = self.pop('tasks')

        dag = DAG()

        if 'config' in self:
            dag._params = DAGConfiguration.from_dict(self['config'])

        clients = self.get('clients')

        if clients:
            init_clients(dag, clients)

        process_tasks(dag, tasks, self, root_path=self._parent_path)

        return dag

    @classmethod
    def auto_load(cls, to_dag=True, starting_dir=None):
        """
        Looks for a pipeline.yaml, generates a DAGSpec and returns a DAG.
        Currently, this is only used by the PloomberContentsManager, this is
        not intended to be a public API since initializing specs from paths
        where we have to recursively look for a pipeline.yaml has some
        considerations regarding relative paths that make this confusing,
        inside the contents manager, all those things are all handled for that
        use case.

        The pipeline.yaml parent folder is temporarily added to sys.path when
        calling DAGSpec.to_dag() to make sure imports work as expected
        """
        path = find_file_recursively('pipeline.yaml',
                                     starting_dir=starting_dir or os.getcwd())

        if path is None:
            if to_dag:
                return None, None, None
            else:
                return None, None

        try:
            spec = cls.from_file(path)

            if to_dag:
                return spec, spec.to_dag(), path
            else:
                return spec, path

        except Exception as e:
            exc = DAGSpecInitializationError('Error initializing DAG from '
                                             'pipeline.yaml')
            raise exc from e

    @classmethod
    def from_file(cls, path):
        """
        Initialize dag spec with yaml file
        """
        with open(str(path)) as f:
            dag_dict = yaml.load(f, Loader=yaml.SafeLoader)

        spec = cls(dag_dict)
        spec._parent_path = str(Path(path).parent)
        return spec


def process_tasks(dag, tasks, dag_spec, root_path=None):
    root_path = root_path or '.'

    # determine if we need to run static analysis
    meta = dag_spec['meta']
    sources = [task_dict['source'] for task_dict in tasks]
    extracted = project.infer_from_path(root_path,
                                        templates=sources,
                                        upstream=meta['extract_upstream'],
                                        product=meta['extract_product'])

    upstream = {}

    for task_dict in tasks:
        source = task_dict['source']

        if dag_spec['meta']['extract_upstream']:
            task_dict['upstream'] = extracted['upstream'][source]

        if dag_spec['meta']['extract_product']:
            task_dict['product'] = extracted['product'][source]

        task, up = task_dict.to_task(dag, root_path)
        upstream[task] = up

    # once we added all tasks, set upstream dependencies
    for task in list(dag.values()):
        for task_dep in upstream[task]:
            task.set_upstream(dag[task_dep])


def init_clients(dag, clients):
    for class_name, dotted_path in clients.items():

        class_ = getattr(tasks, class_name, None)

        if not class_:
            class_ = getattr(products, class_name)

        dag.clients[class_] = load_dotted_path(dotted_path)()


def normalize_task(task):
    if isinstance(task, str):
        return {'source': task}
    else:
        return task



from ploomber.tasks.tasks import (PythonCallable,
                                  ShellScript, DownloadFromURL,
                                  Link, Input)
from ploomber.tasks.Task import Task
from ploomber.tasks.TaskFactory import TaskFactory
from ploomber.tasks.sql import (SQLScript, SQLDump, SQLTransfer,
                                SQLUpload, PostgresCopyFrom)
from ploomber.tasks.notebook import NotebookRunner
from ploomber.tasks.aws import UploadToS3

__all__ = ['PythonCallable', 'ShellScript', 'TaskFactory',
           'Task', 'SQLScript', 'SQLDump', 'SQLTransfer', 'SQLUpload',
           'PostgresCopyFrom', 'NotebookRunner', 'DownloadFromURL',
           'Link', 'Input', 'UploadToS3']



"""
Task abstract class

A Task is a unit of work, it has associated source code and a product
(a persistent object such as a table in a database), it has a name and lives
in a DAG.

[WIP] On subclassing Tasks

Implementation details:

* params (dict), upstream (Param object)
* params vs constructor parameters
* params on render vs params on run
* Implementing Task.run (using the source object, product, TaskBuildError)

Optional:

* Validating PRODUCT_CLASSES_ALLOWED
* Using a client parameter

Task execution model:

Upon instantiation, all tasks start with a WaitingRender status. On successful
render, they move to Skipped if they are up-to-date and all upstream
dependencies are up-to-date as well, WaitingExecution if they
are outdated but upstream dependencies are up-to-date or WaitingUpstream, if
any upstream dependency is outdated.

If render fails, they are marked as ErroredRender and downstream dependencies
are marked as AbortedRender.

When a Task is built, and succeeds, the on_finish hook is executed, if
on_finish succeeds as well, the Task is marked as Executed. If either
the task fails or on_finish fails, the Task is marked as Errored.

When a task is marked as Executed, is saves metadata in its Product,
if the product registered a prepare_metadata hook, the metadata to save
can be modified before is saved.

When a Task is marked as Errored, on_failure hook is executed and downstream
dependencies are marked as Aborted.


TODO: describe BrokenProcesssPool status
"""
import abc
import logging
from datetime import datetime
from ploomber.products import Product, MetaProduct
from ploomber.dag.DAG import DAG
from ploomber.exceptions import TaskBuildError, DAGBuildEarlyStop
from ploomber.tasks.TaskGroup import TaskGroup
from ploomber.constants import TaskStatus
from ploomber.tasks.Upstream import Upstream
from ploomber.tasks.Params import Params
from ploomber.Table import TaskReport, Row
from ploomber.sources.sources import Source
from ploomber.util import isiterable
from ploomber.util.util import callback_check

import humanize


class Task(abc.ABC):
    """
    Abstract class for all Tasks

    Parameters
    ----------
    source: str or pathlib.Path
        Source code for the task, for tasks that do not take source code
        as input (such as PostgresCopyFrom), this can be another thing. The
        source can be a template and can make references to any parameter
        in "params", "upstream" parameters or its own "product", not all
        Tasks have templated source (templating code is mostly used by
        Tasks that take SQL source code as input)
    product: Product
        The product that this task will create upon completion
    dag: DAG
        The DAG holding this task
    name: str
        A name for this task, if None a default will be assigned
    params: dict
        Extra parameters passed to the task on rendering (if templated
        source) or during execution (if not templated source)

    Attributes
    ----------
    params : Params
        A read-only dictionary-like object with params passed, after running
        'product' and 'upstream' are added, if any

    on_render : callable
        Function to execute after rendering. The function can request any of
        the following parmeters: task, client and/or product.

    on_finish : callable
        Function to execute upon execution. Can request the same params as the
        on_render hook.

    on_failure : callable
        Function to execute upon failure. Can request the same params as the
        on_render hook.

    Notes
    -----
    All subclasses must implement the same constuctor to keep the API
    consistent, optional parameters after "params" are ok
    """
    PRODUCT_CLASSES_ALLOWED = None

    @abc.abstractmethod
    def run(self):
        """This is the only required method Task subclasses must implement
        """
        pass

    @abc.abstractmethod
    def _init_source(self, source, kwargs):
        pass

    def __init__(self, source, product, dag, name=None, params=None):
        self._params = Params(params)

        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = self._init_source(source, kwargs)

        if name is None:
            # use name inferred from the source object
            self._name = self._source.name

            if self._name is None:
                raise AttributeError('Task name can only be None if a name'
                                     'can be inferred from the source object. '
                                     'This only works when the task receives a'
                                     'pathlib.Path, when using SourceLoader '
                                     'or in PythonCallable. Pass a value '
                                     'explicitely.')
        else:
            self._name = name

        if dag is None:
            raise TypeError('DAG cannot be None')

        # NOTE: we should get rid of this, maybe just add hooks that are
        # called back on the dag object to avoid having a reference here
        self.dag = dag
        dag._add_task(self)

        if self._source is None:
            raise TypeError('_init_source must return a value, got None')

        if not isinstance(self._source, Source):
            raise TypeError('_init_source must return a subclass of Source')

        if isinstance(product, Product):
            self._product = product

            if self.PRODUCT_CLASSES_ALLOWED is not None:
                if not isinstance(self._product, self.PRODUCT_CLASSES_ALLOWED):
                    raise TypeError('{} only supports the following product '
                                    'classes: {}, got {}'
                                    .format(type(self).__name__,
                                            self.PRODUCT_CLASSES_ALLOWED,
                                            type(self._product).__name__))
        else:
            # if assigned a tuple/list of products, create a MetaProduct
            self._product = MetaProduct(product)

            if self.PRODUCT_CLASSES_ALLOWED is not None:
                if not all(isinstance(p, self.PRODUCT_CLASSES_ALLOWED)
                           for p in self._product):
                    raise TypeError('{} only supports the following product '
                                    'classes: {}, got {}'
                                    .format(type(self).__name__,
                                            self.PRODUCT_CLASSES_ALLOWED,
                                            type(self._product).__name__))

        self._logger = logging.getLogger('{}.{}'.format(__name__,
                                                        type(self).__name__))

        self.product.task = self
        self.client = None

        self.exec_status = TaskStatus.WaitingRender

        self._on_finish = None
        self._on_failure = None
        self._on_render = None

    @property
    def _available_callback_kwargs(self):
        # make it a property so we always get the latest value for self.client
        # given that could be None during init
        return {'task': self,
                'client': self.client,
                'product': self.product}

    @property
    def name(self):
        """
        A str that represents the name of the task, you can access tasks
        in a dag using dag['some_name']
        """
        return self._name

    @property
    def source(self):
        """
        Source is used by the task to compute its output, for most cases
        this is source code, for example PythonCallable takes a function
        as source and SQLScript takes a string with SQL code as source.
        But other tasks might take non-code objects as source, for example,
        PostgresCopyFrom takes a path to a file. If source represents code
        doing str(task.source) will return the string representation
        """
        return self._source

    @property
    def product(self):
        """The product this task will create upon execution
        """
        return self._product

    @property
    def upstream(self):
        """
        A mapping for upstream dependencies {task name} -> [task obkect]
        """
        # this is jus syntactic sugar, upstream relations are tracked by the
        # DAG object

        # this always return a copy to prevent global state if contents
        # are modified (e.g. by using pop)
        return self.dag._get_upstream(self.name)

    @property
    def params(self):
        """
        dict that holds the parameter that will be passed to the task upon
        execution. Before rendering, this will only hold parameters passed
        in the Task constructor. After rendering, this will hold new keys:
        "product" contained the rendered product and "upstream" holding
        upstream parameters if there is any
        """
        return self._params

    @property
    def _lineage(self):
        """
        Set with task names of all the dependencies for this task
        (including dependencies of dependencies)
        """
        # if no upstream deps, there is no lineage
        if not len(self.upstream):
            return None
        else:
            # retrieve lineage: upstream tasks + lineage from upstream tasks
            up = list(self.upstream.keys())
            lineage_up = [up._lineage for up in self.upstream.values() if
                          up._lineage]
            lineage = up + [task for lineage in lineage_up for task in lineage]
            return set(lineage)

    @property
    def on_finish(self):
        """
        Callable to be executed after this task is built successfully
        (passes Task as first parameter)
        """
        return self._on_finish

    @on_finish.setter
    def on_finish(self, value):
        callback_check(value, self._available_callback_kwargs)
        self._on_finish = value

    def _run_on_finish(self):
        # run on finish first, if this fails, we don't want to save metadata
        if self.on_finish:
            kwargs = callback_check(self.on_finish,
                                    self._available_callback_kwargs)
            try:
                self.on_finish(**kwargs)
            except Exception as e:
                msg = ('Exception when running on_finish '
                       'for task "{}": {}'
                       .format(self.name, e))
                self._logger.exception(msg)
                self.exec_status = TaskStatus.Errored
                raise type(e)(msg) from e

        self.product._save_metadata(str(self.source))

        # For most Products, it's ok to do this check before
        # saving metadata, but not for GenericProduct, since the way
        # exists() works is by checking metadata, so we have to do it
        # here, after saving metadata
        if not self.product.exists():
            if isinstance(self.product, MetaProduct):
                raise TaskBuildError(
                    'Error building task "{}": '
                    'the task ran successfully but product '
                    '"{}" does not exist yet '
                    '(task.product.exists() returned False). '
                    .format(self, self.product))
            else:
                raise TaskBuildError(
                    'Error building task "{}": '
                    'the task ran successfully but at least one of the '
                    'products in "{}" does not exist yet '
                    '(task.product.exists() returned False). '
                    .format(self, self.product))

    @property
    def on_failure(self):
        """
        Callable to be executed if task fails (passes Task as first parameter
        and the exception as second parameter)
        """
        return self._on_failure

    @on_failure.setter
    def on_failure(self, value):
        callback_check(value, self._available_callback_kwargs)
        self._on_failure = value

    def _run_on_failure(self):
        if self.on_failure:
            kwargs = callback_check(self.on_failure,
                                    self._available_callback_kwargs)
            self.on_failure(**kwargs)

    @property
    def on_render(self):
        return self._on_render

    @on_render.setter
    def on_render(self, value):
        callback_check(value, self._available_callback_kwargs)
        self._on_render = value

    def _run_on_render(self):
        if self.on_render:
            self._logger.debug('Calling on_render hook on task %s', self.name)

            kwargs = callback_check(self.on_render,
                                    self._available_callback_kwargs)
            try:
                self.on_render(**kwargs)
            except Exception as e:
                msg = ('Exception when running on_render '
                       'for task "{}": {}'
                       .format(self.name, e))
                self._logger.exception(msg)
                self.exec_status = TaskStatus.ErroredRender
                raise type(e)(msg) from e

    @property
    def exec_status(self):
        return self._exec_status

    @exec_status.setter
    def exec_status(self, value):
        # FIXME: this should only be used for th eexecutor to report back
        # status Executed or Errored, reject all other cases, those are handled
        # internally
        if value not in list(TaskStatus):
            raise ValueError('Setting task.exec_status to an unknown '
                             'value: %s', value)

        self._logger.debug('Setting "%s" status to %s', self.name, value)
        self._exec_status = value

        # some executors run this in a subprocess, the main process will not
        # get the new metadata since it was not saved from there, reload
        if value == TaskStatus.Executed:
            self.product.metadata.load()

        # process might crash, propagate now or changes might not be
        # reflected (e.g. if a Task is marked as Aborted, all downtream
        # tasks should be marked as aborted as well)
        # FIXME: this is inefficient, it is better to traverse
        # the dag in topological order but exclude nodes not affected by
        # this change
        self._update_downstream_status()

    def build(self, force=False, catch_exceptions=True):
        """Build a single task

        Although Tasks are primarily designed to execute via DAG.build(), it
        is possible to do so in isolation. However, this only works if the
        task does not have any unrendered upstream dependencies, if that's the
        case, you should call DAG.render() before calling Task.build()

        Examples
        --------
        >>> from pathlib import Path
        >>> from ploomber import DAG
        >>> from ploomber.tasks import PythonCallable
        >>> from ploomber.products import File
        >>> def fn(product):
        ...     Path(str(product)).touch()
        >>> PythonCallable(fn, File('file.txt'), dag=DAG()).build()

        Returns
        -------
        dict
            A dictionary with keys 'run' and 'elapsed'

        Raises
        ------
        TaskBuildError
            If the error failed to build because it has upstream dependencies,
            the build itself failed or build succeded but on_finish hook failed

        DAGBuildEarlyStop
            If any task or on_finish hook raises a DAGBuildEarlyStop error
        """
        if any(t.exec_status == TaskStatus.WaitingRender
               for t in self.upstream.values()):
            raise TaskBuildError('Cannot directly build task "{}" as it '
                                 'has upstream dependencies, call '
                                 'dag.render() first'.format(self.name))

        # This is the public API for uses who'd to run tasks in isolation,
        # we have to make sure we clear product cache status, otherwise
        # this will interfer with other render calls
        self.render(force=force)

        res = self._build(catch_exceptions=catch_exceptions)
        self.product._clear_cached_status()
        return res

    def _build(self, catch_exceptions):
        """
        Private API for building DAGs. This is what executors should call.
        Unlike the public method, this one does not call render, as it
        should happen via a dag.render() call
        """

        if not catch_exceptions:
            res = self._run()
            self._run_on_finish()
            return res
        else:
            try:
                res = self._run()
            except Exception as e:
                msg = 'Error building task "{}"' .format(self.name)
                self._logger.exception(msg)
                self.exec_status = TaskStatus.Errored

                # if there isn't anything left to run, raise exception here
                if self.on_failure is None:
                    if isinstance(e, DAGBuildEarlyStop):
                        raise DAGBuildEarlyStop('Stopping task {} gracefully'
                                                .format(self.name)) from e
                    else:
                        raise TaskBuildError(msg) from e

                build_success = False
                build_exception = e
            else:
                build_success = True
                build_exception = None

            if build_success:
                try:
                    # this not only runs the hook, but also
                    # calls save metadata and checks that the product exists
                    self._run_on_finish()
                except Exception as e:
                    self.exec_status = TaskStatus.Errored
                    msg = ('Exception when running on_finish '
                           'for task "{}": {}'
                           .format(self.name, e))
                    self._logger.exception(msg)

                    if isinstance(e, DAGBuildEarlyStop):
                        raise DAGBuildEarlyStop('Stopping task {} gracefully'
                                                .format(self.name)) from e
                    else:
                        raise TaskBuildError(msg) from e
                else:
                    self.exec_status = TaskStatus.Executed

                return res
            else:
                try:
                    self._run_on_failure()
                except Exception as e:
                    msg = ('Exception when running on_failure '
                           'for task "{}": {}'
                           .format(self.name, e))
                    self._logger.exception(msg)
                    raise TaskBuildError(msg) from e

                if isinstance(build_exception, DAGBuildEarlyStop):
                    raise DAGBuildEarlyStop(
                        'Stopping task {} gracefully'
                        .format(self.name)) from build_exception
                else:
                    msg = 'Error building task "{}"' .format(self.name)
                    raise TaskBuildError(msg) from build_exception

    def _run(self):
        """
        Run task if certain status conditions are ok, otherwise raise a
        TaskBuildError exception
        """
        # cannot keep running, we depend on the render step to get all the
        # parameters resolved (params, upstream, product)
        if self.exec_status == TaskStatus.WaitingRender:
            raise TaskBuildError('Error building task "{}". '
                                 'Cannot build task that has not been '
                                 'rendered, call DAG.render() first'
                                 .format(self.name))

        elif self.exec_status == TaskStatus.Aborted:
            raise TaskBuildError('Attempted to run task "{}", whose '
                                 'status is TaskStatus.Aborted'
                                 .format(self.name))
        elif self.exec_status == TaskStatus.Skipped:
            raise TaskBuildError('Attempted to run task "{}", whose '
                                 'status TaskStatus.Skipped. Render again and'
                                 'set force=True if you want to force '
                                 'execution'
                                 .format(self.name))

        # NOTE: should i fetch metadata here? I need to make sure I have
        # the latest before building
        self._logger.info('Starting execution: %s', repr(self))

        then = datetime.now()
        self.run()
        now = datetime.now()

        elapsed = (now - then).total_seconds()
        self._logger.info('Done. Operation took {:.1f} seconds'
                          .format(elapsed))

        # TODO: also check that the Products were updated:
        # if they did not exist, they must exist now, if they alredy
        # exist, timestamp must be recent equal to the datetime.now()
        # used. maybe run fetch metadata again and validate?

        return TaskReport.with_data(name=self.name, ran=True, elapsed=elapsed)

    def render(self, force=False, outdated_by_code=True):
        """
        Renders code and product, all upstream tasks must have been rendered
        first, for that reason, this method will usually not be called
        directly but via DAG.render(), which renders in the right order.

        Render fully determines whether a task should run or not.

        Parameters
        ----------
        force : bool
            If True, mark status as WaitingExecution/WaitingUpstream even if
            the task is up to, otherwise up to date tasks are marked as
            Skipped

        outdated_by_code : str
            Factors to determine if Task.product is marked outdated when source
            code changes. Otherwise just the upstream timestamps are used.
        """
        self._logger.debug('Calling render on task %s', self.name)

        try:
            self._render_product()
        except Exception as e:
            self.exec_status = TaskStatus.ErroredRender
            raise type(e)('Error rendering product from Task "{}", '
                          ' check the full traceback above for details'
                          .format(repr(self), self.params)) from e

        # Params are read-only for users, but we have to add the product
        # so we do it directly to the dictionary
        self.params._dict['product'] = self.product

        try:
            self.source.render(self.params)
        except Exception as e:
            self.exec_status = TaskStatus.ErroredRender
            raise type(e)('Error rendering source from Task "{}", '
                          ' check the full traceback above for details'
                          .format(repr(self), self.params)) from e

        # Maybe set ._exec_status directly, since no downstream propagation
        # is needed here.
        is_outdated = (self.product
                       ._is_outdated(outdated_by_code=outdated_by_code))

        if not self.upstream:
            if not is_outdated and not force:
                self._exec_status = TaskStatus.Skipped
            else:
                self._exec_status = TaskStatus.WaitingExecution
                self._logger.debug('Forcing status "%s", outdated conditions'
                                   ' ignored...', self.name)
        else:
            all_upstream_done = all([t.exec_status
                                     in {TaskStatus.Executed,
                                         TaskStatus.Skipped}
                                     for t in self.upstream.values()])

            if all_upstream_done and is_outdated:
                self._exec_status = TaskStatus.WaitingExecution
            elif all_upstream_done and not is_outdated and not force:
                self._exec_status = TaskStatus.Skipped
            else:
                self._exec_status = TaskStatus.WaitingUpstream
                self._logger.debug('Forcing status "%s", outdated conditions'
                                   ' ignored...', self.name)

        self._run_on_render()

    def set_upstream(self, other):
        self.dag._add_edge(other, self)

    # FIXME: delete
    def plan(self):
        """Shows a text summary of what this task will execute
        """

        plan = """
        Input parameters: {}
        Product: {}

        Source code:
        {}
        """.format(self.params, self.product, str(self.source))

        print(plan)

    def status(self, return_code_diff=False):
        """Prints the current task status
        """
        p = self.product

        data = {}

        data['name'] = self.name
        data['type'] = type(self).__name__
        data['status'] = self.exec_status.name
        # FIXME: all tasks should have a client property
        data['client'] = (repr(self.client)
                          if hasattr(self, 'client')
                          else None)

        if p.metadata.timestamp is not None:
            dt = datetime.fromtimestamp(p.metadata.timestamp)
            date_h = dt.strftime('%b %d, %y at %H:%M')
            time_h = humanize.naturaltime(dt)
            data['Last updated'] = '{} ({})'.format(time_h, date_h)
        else:
            data['Last updated'] = 'Has not been run'

        outd_data = p._outdated_data_dependencies()
        outd_code = p._outdated_code_dependency()

        outd = False

        if outd_code:
            outd = 'Source code'

        if outd_data:
            if not outd:
                outd = 'Upstream'
            else:
                outd += ' & Upstream'

        data['Outdated?'] = outd

        data['Outdated dependencies'] = outd_data
        data['Outdated code'] = outd_code

        if outd_code and return_code_diff:
            data['Code diff'] = (self.dag
                                 .differ
                                 .get_diff(p.metadata.stored_source_code,
                                           str(self.source),
                                           extension=self.source.extension))
        else:
            outd_code = ''

        data['Product type'] = type(self.product).__name__
        data['Product'] = str(self.product)
        # FIXME: all products should have a client property
        data['Product client'] = (repr(self.product.client)
                                  if hasattr(self.product, 'client')
                                  else None)
        data['Doc (short)'] = _doc_short(self.source.doc)
        data['Location'] = self.source.loc

        return Row(data)

    def to_dict(self):
        """
        Returns a dict representation of the Task, only includes a few
        attributes
        """
        return dict(name=self.name, product=str(self.product),
                    source_code=str(self.source))

    def _render_product(self):
        params_names = list(self.params)

        # add upstream product identifiers to params, if any
        # Params are read-only for users, but we have to add upstream
        # dependencies so we do it directly to the dictionary
        if self.upstream:
            self.params._dict['upstream'] = Upstream({n: t.product for n, t in
                                                      self.upstream.items()},
                                                     name=self.name)

        # render the current product
        try:
            # using the upstream products to define the current product
            # is optional, using the parameters passed in params is also
            # optional
            self.product.render(self.params,
                                optional=set(params_names + ['upstream']))
        except Exception as e:
            raise type(e)('Error rendering Product from Task "{}", '
                          ' check the full traceback above for details'
                          .format(repr(self), self.params)) from e

    def _get_downstream(self):
        # make the _get_downstream more efficient by
        # using the networkx data structure directly
        downstream = []
        for t in self.dag.values():
            if self in t.upstream.values():
                downstream.append(t)
        return downstream

    def _update_downstream_status(self):
        # FIXME: this is inefficient, it is better to traverse
        # the dag in topological order but exclude nodes not affected by
        # this change
        # TODO: move to DAG
        def update_status(task):
            any_upstream_errored_or_aborted = any([t.exec_status
                                                   in (TaskStatus.Errored,
                                                       TaskStatus.Aborted)
                                                   for t
                                                   in task.upstream.values()])
            all_upstream_done = all([t.exec_status
                                     in {TaskStatus.Executed,
                                         TaskStatus.Skipped}
                                     for t in task.upstream.values()])

            if any_upstream_errored_or_aborted:
                task.exec_status = TaskStatus.Aborted
            elif any([t.exec_status in (TaskStatus.ErroredRender,
                                        TaskStatus.AbortedRender)
                      for t in task.upstream.values()]):
                task.exec_status = TaskStatus.AbortedRender
            elif all_upstream_done:
                task.exec_status = TaskStatus.WaitingExecution

        for t in self._get_downstream():
            update_status(t)

    def __rshift__(self, other):
        """ a >> b is the same as b.set_upstream(a)
        """
        other.set_upstream(self)
        # return other so a >> b >> c works
        return other

    def __add__(self, other):
        """ a + b means TaskGroup([a, b])
        """
        if isiterable(other) and not isinstance(other, DAG):
            return TaskGroup([self] + list(other))
        else:
            return TaskGroup((self, other))

    def __repr__(self):
        return ('{}: {} -> {}'
                .format(type(self).__name__, self.name, repr(self.product)))

    def __str__(self):
        return str(self.product)

    def __getstate__(self):
        state = self.__dict__.copy()
        # _logger is not pickable, so we remove them and build
        # them again in __setstate__
        del state['_logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger('{}.{}'.format(__name__,
                                                        type(self).__name__))


def _doc_short(doc):
    if doc is not None:
        return doc.split('\n')[0]
    else:
        return None



import os
from pathlib import Path
import importlib
from functools import wraps, reduce
import base64
import shutil
import inspect
from itertools import chain
from glob import iglob

from ploomber.exceptions import CallbackSignatureError, TaskRenderError


def requires(pkgs, name=None):
    """
    Check if packages were imported, raise ImportError with an appropriate
    message for missing ones
    """
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            missing = [pkg for pkg in pkgs if importlib.util.find_spec(pkg)
                       is None]

            if missing:
                msg = reduce(lambda x, y: x+' '+y, missing)
                fn_name = name or f.__name__

                raise ImportError('{} {} required to use {}. Install {} by '
                                  'running "pip install {}"'
                                  .format(msg,
                                          'is' if len(missing) == 1 else 'are',
                                          fn_name,
                                          'it' if len(
                                              missing) == 1 else 'them',
                                          msg))

            return f(*args, **kwargs)

        return wrapper

    return decorator


@requires(['matplotlib'])
def path2fig(path_to_image, dpi=50):
    # FIXME: having this import at the top causes trouble with the
    # multiprocessing library, moving it here solves the problem but we
    # have to find a better solution.
    # more info: https://stackoverflow.com/q/16254191/709975
    import matplotlib.pyplot as plt

    data = plt.imread(path_to_image)
    height, width, _ = data.shape
    fig = plt.figure()
    fig.set_size_inches((width / dpi, height / dpi))
    ax = plt.Axes(fig, [0, 0, 1, 1])
    ax.set_axis_off()
    fig.add_axes(ax)
    ax.imshow(data)

    return fig


def safe_remove(path):
    if path.exists():
        if path.is_file():
            path.unlink()
        else:
            shutil.rmtree(path)


def image_bytes2html(data):
    fig_base64 = base64.encodebytes(data)
    img = fig_base64.decode("utf-8")
    html = '<img src="data:image/png;base64,' + img + '"></img>'
    return html


# TODO: finish or remove this
# def clean_up_files(dag, interactive=True):
#     """

#     * Get all files generated by the dag
#     * Find the set of parent directories
#     * The parents should only have the files that are generated by tge DAG

#     """
#     # WIP
#     # get products that generate Files
#     paths = [Path(str(t.product)) for t in dag.values()
#              if isinstance(t.product, File)]
#     # each file generates a .source file, also add it
#     paths = [(p, Path(str(p) + '.source')) for p in paths]
#     # flatten list
#     paths = [p for tup in paths for p in tup]

#     # get parents
#     parents = set([p.parent for p in paths])

#     # map parents to its files
#     parents_map = defaultdict(lambda: [])

#     for p in paths:
#         parents_map[str(p.parent)].append(str(p))

#     extra_all = []

#     # for every parent, find the extra files
#     for p in parents:
#         existing = glob(str(p) + '/*')
#         products = parents_map[str(p)]

#         extra = set(existing) - set(products)
#         extra_all.extend(list(extra))

#     for p in extra_all:
#         if interactive:
#             answer = input('Delete {} ? (y/n)'.format(p))

#             if answer == 'y':
#                 safe_remove(p)
#                 print('Deleted {}'.format(p))


def isiterable(obj):
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True


# TODO: add more context to errors, which task and which hook?
def callback_check(fn, available, allow_default=True):
    """
    Check if a callback function signature requests available parameters

    Parameters
    ----------
    fn : callable
        Callable (e.g. a function) to check

    available : dict
        All available params

    allow_default : bool, optional
        Whether allow arguments with default values in "fn" or not

    Returns
    -------
    dict
        Dictionary with requested parameters
    """
    parameters = inspect.signature(fn).parameters
    optional = {name for name, param in parameters.items()
                if param.default != inspect._empty}
    # not all functions have __name__ (e.g. partials)
    fn_name = getattr(fn, '__name__', fn)

    if optional and not allow_default:
        raise CallbackSignatureError('Callback functions cannot have '
                                     'parameters with default values, '
                                     'got: {} in "{}"'.format(optional,
                                                              fn_name))

    required = {name for name, param in parameters.items()
                if param.default == inspect._empty}

    available_set = set(available)
    extra = required - available_set

    if extra:
        raise CallbackSignatureError('Callback function "{}" unknown '
                                     'parameter(s): {}, available ones are: '
                                     '{}'.format(fn_name, extra,
                                                 available_set))

    return {k: v for k, v in available.items() if k in required}


def signature_check(fn, params, task_name):
    """
    Verify if the function signature used as source in a PythonCallable
    task matches available params
    """
    params = set(params)
    parameters = inspect.signature(fn).parameters
    required = {name for name, param in parameters.items()
                if param.default == inspect._empty}

    extra = params - set(parameters.keys())
    missing = set(required) - params

    errors = []

    if extra:
        msg = ('The following params are not part of the function '
               'signature: {}'.format(extra))
        errors.append(msg)

    if missing:
        msg = 'The following params are missing: {}'.format(missing)
        errors.append(msg)

    if extra or missing:
        msg = '. '.join(errors)
        # not all functions have __name__ (e.g. partials)
        fn_name = getattr(fn, '__name__', fn)
        raise TaskRenderError('Error rendering task "{}" initialized with '
                              'function "{}". {}'
                              .format(task_name, fn_name, msg))

    return True


def _parse_module(dotted_path):
    parts = dotted_path.split('.')

    if len(parts) < 2:
        raise ImportError('Invalid module name, must be a dot separated '
                          'string, with at least '
                          '[module_name].[function_name]')

    return '.'.join(parts[:-1]), parts[-1]


def load_dotted_path(dotted_path):
    mod, name = _parse_module(dotted_path)

    try:
        module = importlib.import_module(mod)
    except ImportError as e:
        raise ImportError('An error happened when trying to '
                          'import module "{}"'.format(mod)) from e

    try:
        factory = getattr(module, name)
    except AttributeError as e:
        raise AttributeError('Could not get attribute "{}" from module '
                             '"{}", make sure it is a valid callable'
                             .format(name, mod)) from e

    return factory


def find_file_recursively(name, max_levels_up=6, starting_dir=None):
    """
    Find environment by looking into the current folder and parent folders,
    returns None if no file was found otherwise pathlib.Path to the file
    """
    current_dir = starting_dir or os.getcwd()
    current_dir = Path(current_dir)
    path_to_file = None

    for _ in range(max_levels_up):
        current_path = Path(current_dir, name)

        if current_path.exists():
            path_to_file = current_path.resolve()
            break

        current_dir = current_dir.parent

    return path_to_file



"""
The entry module automatically creates command line interfaces from functions
that return a DAG object.

Example:

Assume module.sub_module.entry_point is a function that returns a DAG.

Run this will parse the function parameters and return them:

    ``$ python -m ploomber.entry module.sub_module.entry_point --help``


If numpydoc is installed, it will parse the docstring (if any) and add it
when using --help

To build the dag returned by the entry point:

    ``$ python -m ploomber.entry module.sub_module.entry_point``


To start an interactive session by loading the entry point first:

    ``$ python -i -m ploomber.entry module.sub_module.entry_point --action status``

``ipython -i -m`` also works. Once the interactive session starts, the object
returned by the entry point will be available in the "dag" variable.

Features:

    * Parse docstring and show it using ``--help``
    * Enable logging to standard output using ``--log LEVEL``
    * Pass function parameters using ``--PARAM_NAME VALUE``
    * If the function is decorated with ``@with_env``, replace env variables using ``--env__KEY VALUE``


``ploomber.entry`` only works with the factory pattern (a function that returns
a dag). But this same pattern can be applied to start interactive sesssions
if you have your own CLI logic (note we are calling the module directly instead
of using ploomber.entry):

    ``$ python -i -m module.sub_module.entry_point --some_arg``


In IPython:

    ``$ ipython -i -m module.sub_module.entry_point -- --some_arg``


Note: you need to add ``--`` to prevent IPython from parsing your custom args

"""
# TODO: print enviornment content on help and maybe on any other command
# it is useful for debugging purposes
import logging
import sys
import importlib
import argparse
import inspect
from pathlib import Path
from collections.abc import Mapping

import yaml

from ploomber.spec.DAGSpec import DAGSpec


def _parse_doc(doc):
    """
    Convert numpydoc docstring to a list of dictionaries
    """
    # no docstring
    if doc is None:
        return {'params': {}, 'summary': None}

    # try to import numpydoc
    docscrape = importlib.import_module('numpydoc.docscrape')

    if not docscrape:
        return {'params': {}, 'summary': None}

    doc = docscrape.NumpyDocString(doc)
    parameters = {p.name: {'desc': ' '.join(p.desc), 'type': p.type}
                  for p in doc['Parameters']}
    summary = doc['Summary']
    return {'params': parameters, 'summary': summary}


def _parse_module(s):
    parts = s.split('.')

    if len(parts) < 2:
        raise ImportError('Invalid module name, must be a dot separated '
                          'string, with at least '
                          '[module_name].[function_name]')

    return '.'.join(parts[:-1]), parts[-1]


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument('entry_point', help='Entry point (DAG)')
    parser.add_argument('--log', help='Enables logging to stdout at the '
                        'specified level', default=None)
    parser.add_argument('--action', help='Action to execute, defaults to '
                        'build', default='build')

    n_positional = len([arg for arg in sys.argv if not arg.startswith('-')])

    if n_positional < 2:
        args = parser.parse_args()
    else:
        # parse entry_point
        entry_point = sys.argv[1]

        if Path(entry_point).exists():
            args = parser.parse_args()

            if args.log is not None:
                logging.basicConfig(level=args.log)

            with open(entry_point) as f:
                dag_dict = yaml.load(f, Loader=yaml.SafeLoader)

            dag = DAGSpec(dag_dict).to_dag()
            getattr(dag, args.action)()
            return dag

        else:

            mod, name = _parse_module(entry_point)

            try:
                module = importlib.import_module(mod)
            except ImportError as e:
                raise ImportError('An error happened when trying to '
                                  'import module "{}"'.format(mod)) from e

            try:
                entry = getattr(module, name)
            except AttributeError as e:
                raise AttributeError('Could not get attribute "{}" from module '
                                     '"{}", make sure it is a valid callable'
                                     .format(name, mod)) from e

            doc = _parse_doc(entry.__doc__)

            def get_desc(arg):
                arg_data = doc['params'].get(arg)
                return None if arg_data is None else arg_data['desc']

            sig = inspect.signature(entry)

            defaults = {k: v.default for k, v in sig.parameters.items()
                        if v.default != inspect._empty}
            required = [k for k, v in sig.parameters.items()
                        if v.default == inspect._empty]

            for arg, default in defaults.items():
                parser.add_argument('--'+arg,
                                    help=get_desc(arg))

            for arg in required:
                parser.add_argument(arg, help=get_desc(arg))

            # if entry point was decorated with @with_env, add arguments
            # to replace declared variables in env.yaml
            if hasattr(entry, '_env_dict'):
                flat_env_dict = _flatten_dict(entry._env_dict._data)
                for arg, val in flat_env_dict.items():
                    parser.add_argument('--env__'+arg,
                                        help='Default: {}'.format(val))

            args = parser.parse_args()

            if args.log is not None:
                logging.basicConfig(level=args.log)

            # required by the function signature
            kwargs = {key: getattr(args, key) for key in required}

            # env and function defaults replaced
            replaced = {name: getattr(args, name)
                        for name in dir(args)
                        if not name.startswith('_')
                        if getattr(args, name) is not None
                        if name not in {'entry_point', 'action', 'log'}}

            # TODO: add a way of test this by the parameters it will use to
            # call the function, have an aux function to get those then another
            # to execute, test using the first one
            obj = entry(**{**kwargs, **replaced})

            print(getattr(obj, args.action)())

            return obj


def _flatten_dict(d, prefix=''):
    """
    Convert a nested dict: {'a': {'b': 1}} -> {'a__b': 1}
    """
    out = {}

    for k, v in d.items():
        if isinstance(v, Mapping):
            out = {**out, **_flatten_dict(v, prefix=prefix + k + '__')}
        else:
            out[prefix+k] = v

    return out



from jinja2 import Environment, Template
from jinja2.nodes import Assign
from ploomber import products


class JinjaUpstreamIntrospector:
    def __init__(self):
        self.keys = []

    def __getitem__(self, key):
        self.keys.append(key)


def extract_upstream_from_sql(sql):
    """Extract upstream keys used in a templated SQL script
    """
    upstream = JinjaUpstreamIntrospector()
    params = {'upstream': upstream}

    product = extract_product_from_sql(sql)

    # FIXME: we need to pass the class if the product is declared there and
    # want to call render, but this is inefficient, find a better way
    if product:
        params[type(product).__name__] = type(product)

    Template(sql).render(params)
    return set(upstream.keys) if len(upstream.keys) else None


def extract_product_from_sql(sql):
    """
    Extract an object from a SQL template that defines as product variable:

    {% set product = SOME_CLASS(...) %}

    Where SOME_CLASS is a class defined in ploomber.products. If no product
    variable is defined, returns None
    """
    env = Environment()
    ast = env.parse(sql)
    variables = {n.target.name: n.node for n in ast.find_all(Assign)}

    if 'product' not in variables:
        return None
    else:
        product = variables['product']
        # TODO: check product.node.ctx == 'load'
        class_ = getattr(products, product.node.name)
        arg = product.args[0].as_const()
        return class_(arg)



import importlib
import platform
from pathlib import Path
from collections.abc import Mapping

import yaml

from ploomber.env import validate
from ploomber.env.expand import EnvironmentExpander
from ploomber.env.FrozenJSON import FrozenJSON
from ploomber.util.util import find_file_recursively


# TODO: custom expanders, this could be done trough another special directive
# such as _expander_class to know which class to use
class EnvDict(Mapping):
    """
    Implements the initialization functionality for Env, except it allows
    to more than one instance to exist, this is used internally to allow
    factory functions introspection without having to create an actual Env

    Parameters
    ----------
    source : dict or str
        If str, it will be interpreted as a path to a YAML file

    """

    def __init__(self, source):

        # load data
        (self._raw_data,
         # this will be None if source is a dict
         self.path_to_env) = load_from_source(source)

        # check raw data is ok
        validate.raw_data_keys(self._raw_data)

        # expand _module special key, return its expanded value
        self.preprocessed = raw_preprocess(self._raw_data,
                                           self.path_to_env)

        # initialize expander, which converts placeholders to their values
        # we need to pass path_to_env since the {{here}} placeholder resolves
        # to its parent
        self.expander = EnvironmentExpander(self.preprocessed,
                                            self.path_to_env)

        # now expand all values
        self._data = self.expander.expand_raw_dictionary(self._raw_data)

    def __getattr__(self, key):
        error = AttributeError("'{}' object has no attribute '{}'"
                               .format(type(self).__name__, key))
        # do not look up special atttributes this way!
        if key.startswith('__') and key.endswith('__'):
            raise error

        if key in self:
            return self[key]
        else:
            raise error

    def __getitem__(self, key):
        if key in self.preprocessed:
            return FrozenJSON(self.preprocessed[key])
        else:
            return FrozenJSON(self._data[key])

    def __iter__(self):
        for k in self._data:
            yield k

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, str(self))

    def _replace_value(self, value, keys_all):
        keys_to_final_dict = keys_all[:-1]
        key_to_edit = keys_all[-1]

        dict_to_edit = self._data

        for e in keys_to_final_dict:
            dict_to_edit = dict_to_edit[e]

        if dict_to_edit.get(key_to_edit) is None:
            dotted_path = '.'.join(keys_all)
            raise KeyError('Trying to replace key "{}" in env, '
                           'but it does not exist'
                           .format(dotted_path))

        dict_to_edit[key_to_edit] = (self.expander
                                     .expand_raw_value(value,
                                                       keys_all))


def load_from_source(source):
    """
    Loads from a dictionary or a YAML and applies preprocesssing to the
    dictionary

    Returns
    -------
    dict
        Raw dictioanry
    pathlib.Path
        Path to the loaded file, None if source is a dict
    str
        Name, if loaded from a YAML file with the env.{name}.yaml format,
        None if another format or if source is a dict
    """
    if isinstance(source, Mapping):
        # dictiionary, path
        return source, None

    elif source is None:
        # look for an env.{name}.yaml, if that fails, try env.yaml
        name = platform.node()
        path_found = find_env_w_name(name)

        if path_found is None:
            raise FileNotFoundError('Tried to initialize environment with '
                                    'None, but automatic '
                                    'file search failed to locate '
                                    'env.{}.yaml nor env.yaml in the '
                                    'current directory nor 6 levels up'
                                    .format(name))
        else:
            source = path_found

    elif isinstance(source, (str, Path)):
        # if not pointing to a file, try to locate it...
        if not Path(source).exists():
            source_found = find_file_recursively(source)

            if source_found is None:
                raise FileNotFoundError('Could not find file "{}" in the '
                                        'current working directory nor '
                                        '6 levels up'.format(source))
            else:
                source = source_found

    with open(str(source)) as f:
        try:
            raw = yaml.load(f, Loader=yaml.SafeLoader)
        except Exception as e:
            raise type(e)('yaml.load failed to parse your YAML file '
                          'fix syntax errors and try again') from e

    path = Path(source).resolve()

    return raw, path


def raw_preprocess(raw, path_to_raw):
    """
    Preprocess a raw dictionary. If a '_module' key exists, it
    will be expanded: first, try to locate a module with that name and resolve
    to their location (root __init__.py parent), if no module is found,
    interpret as a path to the project's root folder, checks that the folder
    actually exists. '{{here}}' is also allowed, which resolves to the
    path_to_raw, raises Exception if path_to_raw is None

    Returns
    -------
    preprocessed : dict
        Dict with preprocessed keys (empty dictionary it no special
        keys exist in raw)

    Parameters
    ----------
    raw : dict
        Raw data dictionary
    path_to_raw : str
        Path to file where dict was read from, it read from a dict, pass None
    """
    module = raw.get('_module')
    preprocessed = {}

    if module:

        if raw['_module'] == '{{here}}':

            if path_to_raw is not None:
                preprocessed['_module'] = path_to_raw.parent
            else:
                raise ValueError('_module cannot be {{here}} if '
                                 'not loaded from a file')
        else:
            try:
                module_spec = importlib.util.find_spec(module)
            except ValueError:
                # raises ValueError if passed "."
                module_spec = None

            if not module_spec:
                path_to_module = Path(module)

                if not (path_to_module.exists()
                        and path_to_module.is_dir()):
                    raise ValueError('Could not resolve _module "{}", '
                                     'failed to import as a module '
                                     'and is not a directory'
                                     .format(module))

            else:
                path_to_module = Path(module_spec.origin).parent

            preprocessed['_module'] = path_to_module

    return preprocessed


def find_env_w_name(name):
    """
    Find environment named 'env.{name}.yaml' by looking into the current
    directory and upper folders. If this fails, attempt to do the same
    for a file named just 'env.yaml'

    Returns
    -------
    path_to_env : pathlib.Path
        Path to environment file, None if no file could be found
    """
    path = find_file_recursively(name='env.{}.yaml'.format(name))

    if path is None:
        return find_file_recursively(name='env.yaml')
    else:
        return path



"""
Testing SQL relations
"""
from jinja2 import Template


def nulls_in_columns(client, cols, product):
    """Check if any column has NULL values, returns bool
    """
    sql = Template("""
    SELECT EXISTS(
        SELECT * FROM {{product}}
        WHERE {{cols | join(' is null or ') }} is null
    )
    """).render(cols=cols, product=product)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = bool(cur.fetchone()[0])
    cur.close()

    return output


def distinct_values_in_column(client, col, product):
    """Get distinct values in a column, returns a set
    """
    sql = Template("""
    SELECT DISTINCT {{col}} FROM {{product}}
    """).render(col=col, product=product)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = cur.fetchall()
    cur.close()

    return set(o[0] for o in output)


def duplicates_in_column(client, col, product):
    """Check if a column has duplicated values, returns bool
    """
    sql = Template("""
    SELECT EXISTS(
        SELECT {{col}}, COUNT(*)
        FROM {{product}}
        GROUP BY {{col}}
        HAVING COUNT(*) > 1
    )
    """).render(col=col, product=product)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = bool(cur.fetchone()[0])
    cur.close()

    return output


def range_in_column(client, col, product):
    """Get range for a column, returns a (min_value, max_value) tuple
    """
    sql = Template("""
    SELECT MIN({{col}}), MAX({{col}}) FROM {{product}}
    """).render(col=col, product=product)

    cur = client.connection.cursor()
    cur.execute(sql)
    output = cur.fetchone()
    cur.close()

    return output



import jupytext
from collections.abc import Mapping
import parso

from ploomber.sources.NotebookSource import _find_cell_with_tag


def extract_variable_from_parameters(code_str, fmt, variable):
    """
    Infer upstream dependencies from a .py or .ipynb source string.
    Must contain a "parameters" cell
    """
    nb = jupytext.reads(code_str, fmt)
    cell, _ = _find_cell_with_tag(nb, 'parameters')

    if cell is None:
        raise ValueError('Notebook does not have a cel with the tag '
                         '"parameters"')

    if variable == 'upstream':
        return infer_dependencies_from_code_cell(cell.source)
    elif variable == 'product':
        return get_product(cell.source)
    else:
        raise ValueError('variable must be upstream or product')


def extract_variable_from_code_cell(code_str, name):
    variable_found = False
    value = None

    p = parso.parse(code_str)

    for ch in p.children:
        # FIXME: this works but we should find out what's the difference
        # between these two and if they are the only two valid cases
        if ch.type in ['simple_stmt', 'expr_stmt']:
            if ch.type == 'simple_stmt':
                stmt = ch.children[0]
            elif ch.type == 'expr_stmt':
                stmt = ch

            defined = stmt.get_defined_names()

            if len(defined) == 1 and defined[0].value == name:
                variable_found = True
                value = eval(stmt.children[2].get_code())

    return variable_found, value


def get_product(code_str):
    product_found, product = extract_variable_from_code_cell(code_str,
                                                             'product')

    if not product_found:
        raise ValueError("Could not parse a valid 'product' variable "
                         "from code:\n%s" % code_str)
    else:
        return product


def infer_dependencies_from_code_cell(code_str):
    """
    Infer dependencies from a single Python cell. Looks for a cell that
    defines an upstream variable which must be either a dictionary or None
    """
    upstream_found, upstream = extract_variable_from_code_cell(code_str,
                                                               'upstream')

    if not upstream_found:
        raise ValueError("Could not parse a valid 'upstream' variable "
                         "from code:\n'%s'. If the notebook "
                         "does not have dependencies add "
                         "upstream = None" % code_str)
    else:
        valid_types = (Mapping, list, tuple, set)
        if not (isinstance(upstream, valid_types) or upstream is None):
            raise ValueError("Found an upstream variable but it is not a "
                             "valid type (dictionary, list, tuple set or None "
                             ", got '%s' type from code:\n"
                             "'%s'" % (type(upstream), code_str))
        elif isinstance(upstream, valid_types):
            return set(upstream)
        else:
            return None



"""
Products are persistent changes triggered by Tasks such as a new file
in the local filesystem or a table in a database
"""
import warnings
import json
import shutil
import os
from pathlib import Path
from ploomber.products.Product import Product
from ploomber.placeholders.Placeholder import Placeholder


class File(Product):
    """A file (or directory) in the local filesystem

    Parameters
    ----------
    identifier: str or pathlib.Path
        The path to the file (or directory), can contain placeholders
        (e.g. {{placeholder}})
    """

    def _init_identifier(self, identifier):
        if not isinstance(identifier, (str, Path)):
            raise TypeError('File must be initialized with a str or a '
                            'pathlib.Path')

        return Placeholder(str(identifier))

    @property
    def _path_to_file(self):
        return Path(str(self._identifier))

    # TODO: rename this to path_to_metadata?
    @property
    def _path_to_stored_source_code(self):
        return Path(str(self._path_to_file) + '.source')

    def fetch_metadata(self):
        empty = dict(timestamp=None, stored_source_code=None)
        # but we have no control over the stored code, it might be missing
        # so we check, we also require the file to exists: even if the
        # .source file exists, missing the actual data file means something
        # if wrong anf the task should run again
        if (self._path_to_stored_source_code.exists()
                and self._path_to_file.exists()):
            content = self._path_to_stored_source_code.read_text()

            try:
                parsed = json.loads(content)
            except json.JSONDecodeError:
                # for compatibility with the previous version
                stored_source_code = content
                timestamp = self._path_to_stored_source_code.stat().st_mtime

                warnings.warn('Migrating metadata to new format...')
                loaded = dict(timestamp=timestamp,
                              stored_source_code=stored_source_code)
                self.save_metadata(loaded)
                return loaded
            else:
                # TODO: validate 'stored_source_code', 'timestamp' exist
                return parsed
        else:
            return empty

    def save_metadata(self, metadata):
        (self._path_to_stored_source_code
            .write_text(json.dumps(metadata)))

    def exists(self):
        return self._path_to_file.exists()

    def delete(self, force=False):
        # force is not used for this product but it is left for API
        # compatibility
        if self.exists():
            self.logger.debug('Deleting %s', self._path_to_file)
            if self._path_to_file.is_dir():
                shutil.rmtree(str(self._path_to_file))
            else:
                os.remove(str(self._path_to_file))
        else:
            self.logger.debug('%s does not exist ignoring...',
                              self._path_to_file)

    @property
    def suffix(self):
        return Path(self._identifier.best_str(shorten=False)).suffix



import sqlite3
import json

from jinja2 import Template

from ploomber.products import Product
from ploomber.products.serializers import Base64Serializer
from ploomber.placeholders.Placeholder import SQLRelationPlaceholder


class ProductWithClientMixin:
    """
    Adds the client property to a Product with the hierarchical resolution
    logic: Product -> Task -> DAG.clients. Product.client is only used
    for storing metadata
    """
    @property
    def client(self):
        # FIXME: this nested reference looks ugly, how can we improve this?
        if self._client is None:
            if self._task is None:
                raise ValueError('Cannot obtain client for this product, '
                                 'the constructor did not receive a client '
                                 'and this product has not been assigned '
                                 'to a DAG yet, hence we cannot look up '
                                 'dag.clients')

            default = self.task.dag.clients.get(type(self))

            if default is None:
                raise ValueError('{} must be initialized with a client'
                                 .format(type(self).__name__))
            else:
                self._client = default

        return self._client


class SQLiteBackedProductMixin(ProductWithClientMixin):
    def _create_metadata_relation(self):
        create_metadata = """
        CREATE TABLE IF NOT EXISTS _metadata (
            name TEXT NOT NULL,
            schema TEXT,
            metadata BLOB,
            UNIQUE(schema, name)
        )
        """
        self.client.execute(create_metadata)

    def _get_schema(self):
        # the actual implementation can be a simple identifier or a sql
        # identifier
        if hasattr(self, 'schema'):
            # there is a schema (can be None)
            return self.schema
        else:
            # this means there is no schema
            return False

    def fetch_metadata(self):
        self._create_metadata_relation()

        query = Template("""
        SELECT metadata FROM _metadata
        WHERE name = '{{name}}'
        {% if schema is not sameas false %}
            {% if schema %}
                AND schema = '{{schema}}'
            {% else %}
                AND schema IS NULL
            {% endif %}
        {% endif %}
        """).render(name=self.name, schema=self._get_schema())

        cur = self.client.connection.cursor()
        cur.execute(query)
        records = cur.fetchone()
        cur.close()

        if records:
            metadata_bin = records[0]
            return json.loads(metadata_bin.decode("utf-8"))
        else:
            return None

    def save_metadata(self, metadata):
        self._create_metadata_relation()

        metadata_bin = json.dumps(metadata).encode('utf-8')
        cur = self.client.connection.cursor()
        schema = self._get_schema()

        if schema is not False:
            if schema is None:
                cur.execute('DELETE FROM _metadata WHERE name = ?',
                            (self.name,))
            else:
                cur.execute('DELETE FROM _metadata '
                            'WHERE name = ? AND schema = ?',
                            (self.name, schema))

            # we cannot rely on INSERT INTO since NULL schema values are
            # allowed
            query = """
                INSERT INTO _metadata (metadata, name, schema)
                VALUES(?, ?, ?)
            """
            cur.execute(query, (sqlite3.Binary(metadata_bin),
                                self.name,
                                schema))
        else:
            query = """
                REPLACE INTO _metadata(metadata, name)
                VALUES(?, ?)
            """
            cur.execute(query, (sqlite3.Binary(metadata_bin),
                                self.name))

        self.client.connection.commit()
        cur.close()

    def _delete_metadata(self):
        self._create_metadata_relation()

        query = Template("""
        DELETE FROM _metadata
        WHERE name = '{{name}}'
        {% if schema is not sameas false %}
            {% if schema %}
                AND schema = '{{schema}}'
            {% else %}
                AND schema IS NULL
            {% endif %}
        {% endif %}
        """).render(name=self.name, schema=self._get_schema())

        cur = self.client.connection.cursor()
        cur.execute(query)
        cur.close()


class SQLiteRelation(SQLiteBackedProductMixin, Product):
    """A SQLite relation

    Parameters
    ----------
    identifier: tuple of length 3 or 2
        A tuple with (schema, name, kind) where kind must be either 'table'
        or 'view'. If passed a tuple with length 2, schema is assumed None.
        Schemas in SQLite represent other databases when using the ATTACH
        command.
    client: ploomber.clients.DBAPIClient or SQLAlchemyClient, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]


    Examples
    --------
    >>> from ploomber.products import SQLiteRelation
    >>> relation = SQLiteRelation(('schema', 'some_table', 'table'))
    >>> str(relation) # returns qualified name
    """

    def __init__(self, identifier, client=None):
        super().__init__(identifier)
        self._client = client

    def _init_identifier(self, identifier):
        return SQLRelationPlaceholder(identifier)

    def exists(self):
        query = """
        SELECT name
        FROM sqlite_master
        WHERE type = '{kind}'
        AND name = '{name}'
        """.format(kind=self.kind,
                   name=self.name)

        cur = self.client.connection.cursor()
        cur.execute(query)
        exists = cur.fetchone() is not None
        cur.close()
        return exists

    def delete(self):
        """Deletes the product
        """
        query = ("DROP {kind} IF EXISTS {relation}"
                 .format(kind=self.kind,
                         relation=str(self)))
        self.logger.debug('Running "{query}" on the databse...'
                          .format(query=query))
        self.client.execute(query)

    @property
    def name(self):
        return self._identifier.name

    @property
    def schema(self):
        return self._identifier.schema

    @property
    def kind(self):
        return self._identifier.kind


# FIXME: self._identifier should not be accessed direclty since it might
# be a placeholder
class PostgresRelation(ProductWithClientMixin, Product):
    """A PostgreSQL relation

    Parameters
    ----------
    identifier: tuple of length 3
        A tuple with (schema, name, kind) where kind must be either 'table'
        or 'view'
    client: ploomber.clients.DBAPIClient or SQLAlchemyClient, optional
        The client used to connect to the database. Only required
        if no dag-level client has been declared using dag.clients[class]

    Examples
    --------
    >>> from ploomber.products import PostgresRelation
    >>> relation = PostgresRelation(('schema', 'some_table', 'table'))
    >>> str(relation) # returns qualified name
    """
    # FIXME: identifier has schema as optional but that introduces ambiguity
    # when fetching metadata and checking if the table exists so maybe it
    # should be required

    def __init__(self, identifier, client=None):
        self._client = client
        super().__init__(identifier)

    def _init_identifier(self, identifier):
        return SQLRelationPlaceholder(identifier)

    def fetch_metadata(self):
        cur = self.client.connection.cursor()

        if self.schema:
            schema = self.schema
        else:
            # if schema is empty, we have to find out the default one
            query = """
            SELECT current_schema()
            """
            cur.execute(query)
            schema = cur.fetchone()[0]

        # https://stackoverflow.com/a/11494353/709975
        query = """
        SELECT description
        FROM pg_description
        JOIN pg_class ON pg_description.objoid = pg_class.oid
        JOIN pg_namespace ON pg_class.relnamespace = pg_namespace.oid
        WHERE nspname = %(schema)s
        AND relname = %(name)s
        """
        cur.execute(query, dict(schema=schema, name=self.name))
        metadata = cur.fetchone()

        cur.close()

        # no metadata saved
        if metadata is None:
            return None
        else:
            return Base64Serializer.deserialize(metadata[0])

        # TODO: also check if metadata  does not give any parsing errors,
        # if yes, also return a dict with None values, and maybe emit a warn

    def save_metadata(self, metadata):
        metadata = Base64Serializer.serialize(metadata)

        query = (("COMMENT ON {} {} IS '{}';"
                  .format(self.kind,
                          self._identifier,
                          metadata)))

        cur = self.client.connection.cursor()
        cur.execute(query)
        self.client.connection.commit()
        cur.close()

    def exists(self):
        cur = self.client.connection.cursor()

        if self.schema:
            schema = self.schema
        else:
            # if schema is empty, we have to find out the default one
            query = """
            SELECT current_schema()
            """
            cur.execute(query)
            schema = cur.fetchone()[0]

        # https://stackoverflow.com/a/24089729/709975
        query = """
        SELECT EXISTS (
            SELECT 1
            FROM   pg_catalog.pg_class c
            JOIN   pg_catalog.pg_namespace n ON n.oid = c.relnamespace
            WHERE  n.nspname = %(schema)s
            AND    c.relname = %(name)s
        );
        """

        cur.execute(query, dict(schema=schema,
                                name=self.name))
        exists = cur.fetchone()[0]
        cur.close()
        return exists

    def delete(self, force=False):
        """Deletes the product
        """
        cascade = 'CASCADE' if force else ''
        query = ("DROP {} IF EXISTS {} {}"
                 .format(self.kind, self, cascade))
        self.logger.debug('Running "%s" on the databse...', query)

        cur = self.client.connection.cursor()
        cur.execute(query)
        cur.close()
        self.client.connection.commit()

    @property
    def name(self):
        return self._identifier.name

    @property
    def schema(self):
        return self._identifier.schema

    @property
    def kind(self):
        return self._identifier.kind



from ploomber.products.Product import Product
from ploomber.products.MetaProduct import MetaProduct
from ploomber.products.File import File
from ploomber.products.GenericProduct import (GenericProduct,
                                              GenericSQLRelation,
                                              SQLRelation)
from ploomber.products.sql import SQLiteRelation, PostgresRelation

__all__ = ['File', 'MetaProduct', 'Product', 'SQLiteRelation',
           'PostgresRelation', 'GenericProduct', 'GenericSQLRelation',
           'SQLRelation']



"""
Documentation: https://ploomber.readthedocs.io/en/stable/modules/clients.html
"""
from ploomber.clients import SQLAlchemyClient


def get_client():
    """Return a client to the database
    """
    # URI documentation: https://docs.sqlalchemy.org/en/13/core/engines.html
    # NOTE: do not expose database details in your source code, use
    # a secure credentials storage (check out the keyring package), other
    # (less secure) options are environment variables and a file in a standard
    # location (set read permission only for the current user)
    return SQLAlchemyClient('dialect+driver://user:pwd@host:port/database')



import re
import ast
import pydoc
import getpass
from copy import deepcopy, copy
from collections.abc import Mapping
from pathlib import Path

from jinja2 import Template

from ploomber.placeholders import util
from ploomber import repo


class EnvironmentExpander:
    """
    Conver values in the raw dictionary by expanding tags such as {{git}},
    {{version}} or {{here}}. See `expand_raw_value` for more details
    """

    def __init__(self, preprocessed, path_to_env=None,
                 version_requires_import=False):
        self._preprocessed = preprocessed

        # {{here}} resolves to this value
        self._path_to_here = (None if path_to_env is None
                              else str(Path(path_to_env).parent))
        # we compute every placeholder's value so we only do it once
        self._placeholders = {}

        self._version_requires_import = version_requires_import

    def expand_raw_dictionary(self, raw):
        data = deepcopy(raw)

        for (d, current_key,
             current_val, parent_keys) in iterate_nested_dict(data):
            d[current_key] = self.expand_raw_value(current_val, parent_keys)

        return data

    def expand_raw_value(self, raw_value, parents):
        """
        Expand a string with placeholders

        Parameters
        ----------
        raw_value : str
            The original value to expand
        parents : list
            The list of parents to get to this value in the dictionary

        Notes
        -----
        If for a given raw_value, the first parent is 'path', expanded value
        is casted to pathlib.Path object and .expanduser() is called,
        furthermore, if raw_value ends with '/', a directory is created if
        it does not currently exist
        """
        placeholders = util.get_tags_in_str(raw_value)

        if not placeholders:
            value = raw_value
        else:
            # get all required placeholders
            params = {k: self.load_placeholder(k) for k in placeholders}
            value = Template(raw_value).render(**params)

        if parents:
            if parents[0] == 'path':

                # value is a str (since it was loaded from a yaml file),
                # if it has an explicit trailing slash, interpret it as
                # a directory and create it, we have to do it at this point,
                # because once we cast to Path, we lose the trailing slash
                if value.endswith('/'):
                    self._try_create_dir(value)

                return Path(value).expanduser()
            else:
                return value

    def _try_create_dir(self, value):
        # make sure to expand user to avoid creating a "~" folder
        path = Path(value).expanduser()

        if not path.exists():
            path.mkdir(parents=True)

    def load_placeholder(self, key):
        if key not in self._placeholders:
            if hasattr(self, 'get_'+key):
                self._placeholders[key] = getattr(self, 'get_'+key)()
            else:
                raise RuntimeError('Unknown placeholder "{}"'.format(key))

        return self._placeholders[key]

    def _get_version_importing(self):
        module_path = self._preprocessed.get('_module')

        if not module_path:
            raise KeyError('_module key is required to use version '
                           'placeholder')

        # is this ok to do? /path/to/{module_name}
        module_name = str(Path(module_path).name)
        module = pydoc.locate(module_name)

        if module is None:
            raise ImportError('Unabe to import module with name "{}"'
                              .format(module_name))

        if hasattr(module, '__version__'):
            return module.__version__
        else:
            raise RuntimeError('Module "{}" does not have a __version__ '
                               'attribute '.format(module))

    def _get_version_without_importing(self):
        if '_module' not in self._preprocessed:
            raise KeyError('_module key is required to use version '
                           'placeholder')

        content = (self._preprocessed['_module'] / '__init__.py').read_text()

        version_re = re.compile(r'__version__\s+=\s+(.*)')

        version = str(ast.literal_eval(version_re.search(
                      content).group(1)))
        return version

    def get_version(self):
        if self._version_requires_import:
            return self._get_version_importing()
        else:
            return self._get_version_without_importing()

    def get_user(self):
        return getpass.getuser()

    def get_here(self):
        if self._path_to_here:
            return self._path_to_here
        else:
            raise RuntimeError('here placeholder is only available '
                               'when env was initialized from a file')

    def get_git(self):
        module = self._preprocessed.get('_module')

        if not module:
            raise KeyError('_module key is required to use git placeholder')

        return repo.get_env_metadata(module)['git_location']


def iterate_nested_dict(d, preffix=[]):
    """
    Iterate over all values (possibly nested) in a dictionary

    Yields: dict holding the value, current key, current value, list of keys
    to get to this value
    """
    for k, v in d.items():
        if isinstance(v, Mapping):
            preffix_new = copy(preffix)
            preffix_new.append(k)
            for i in iterate_nested_dict(v, preffix_new):
                yield i
        else:
            preffix_new = copy(preffix)
            preffix_new.append(k)
            yield d, k, v, preffix_new



from pathlib import Path
import tempfile
import inspect

import jupyter_client
import papermill
import parso
import nbformat


class CallableDebugger:
    """

    Examples
    --------
    >>> CallableDebugger(fn, {'param': 1})
    >>> tmp = debugger._to_nb()
    >>> tmp # modify tmp notebook
    >>> debugger._overwrite_from_nb(tmp)
    """

    def __init__(self, fn, params):
        self.fn = fn
        self.file = inspect.getsourcefile(fn)
        # how do we deal with def taking more than one line?
        lines, start = inspect.getsourcelines(fn)
        self.lines = (start, start + len(lines))
        self.params = params
        _, self.tmp_path = tempfile.mkstemp(suffix='.ipynb')
        self.body_start = None

    def _to_nb(self):
        """
        Returns the function's body in a notebook (tmp location), insert
        injected parameters
        """
        body_elements, self.body_start = parse(self.fn)
        body_to_nb(body_elements, self.tmp_path)

        # TODO: inject cells with imports + functions + classes defined
        # in the file (should make this the cwd for imports in non-packages
        # to work though) - if i do that, then I should probably do the same
        # for notebook runner to be consistent

        papermill.execute_notebook(self.tmp_path, self.tmp_path,
                                   prepare_only=True,
                                   parameters=self.params)

        return self.tmp_path

    def _overwrite_from_nb(self, path):
        """
        Overwrite the function's body with the notebook contents, excluding
        injected parameters and cells whose first line is #
        """
        # add leading space
        nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)

        code_cells = [c['source'] for c in nb.cells if c['cell_type'] == 'code'
                      and 'injected-parameters'
                      not in c['metadata'].get('tags', [])
                      and c['source'][:2] != '#\n']

        # add 4 spaces to each code cell, exclude white space lines
        code_cells = [indent_cell(code) for code in code_cells]

        content = Path(self.file).read_text().splitlines()

        # first element: where the function starts + offset due to signature
        # line break
        # second: new body
        # third: the rest of the file
        keep_until = self.lines[0] + self.body_start
        new_content = (content[:keep_until]
                       + code_cells + ['\n'] + content[self.lines[1]:])

        Path(self.file).write_text('\n'.join(new_content))

    def __enter__(self):
        self.tmp_path = self._to_nb()
        return self.tmp_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._overwrite_from_nb(self.tmp_path)
        Path(self.tmp_path).unlink()

    def __del__(self):
        tmp = Path(self.tmp_path)
        if tmp.exists():
            tmp.unlink()


def indent_line(l):
    return '    ' + l if l else ''


def indent_cell(code):
    return '\n'.join([indent_line(l) for l in code.splitlines()])


def parse(fn):
    # TODO: exclude return at the end, what if we find more than one?
    # maybe do not support functions with return statements for now
    # getsource adds a new line at the end of the the function, we don't need
    # this
    s = inspect.getsource(fn).rstrip()
    module = parso.parse(s)
    body = module.children[0].children[-1]

    # parso is adding a new line as first element, not sure if this
    # happens always though
    if isinstance(body.children[0], parso.python.tree.Newline):
        body_elements = body.children[1:]
    else:
        body_elements = body.children

    # return body and the last line of the signature
    return body_elements, body.start_pos[0] - 1


def body_to_nb(body_elements, path):
    """
    Converts a Python function to a notebook
    """
    nb_format = nbformat.versions[nbformat.current_nbformat]
    nb = nb_format.new_notebook()

    for statement in body_elements:
        # parso incluses new line tokens, remove any trailing whitespace
        lines = [l[4:] for l in statement.get_code().rstrip().split('\n')]
        cell = nb_format.new_code_cell(source='\n'.join(lines))
        nb.cells.append(cell)

    k = jupyter_client.kernelspec.get_kernel_spec('python3')

    nb.metadata.kernelspec = {
        "display_name": k.display_name,
        "language": k.language,
        "name": 'python3'
    }

    nbformat.write(nb, path)



from ploomber.sources.sources import (SQLScriptSource,
                                      SQLQuerySource,
                                      GenericSource,
                                      FileSource,
                                      Source, EmptySource)
from ploomber.sources.NotebookSource import NotebookSource
from ploomber.sources.PythonCallableSource import PythonCallableSource

__all__ = ['PythonCallableSource', 'SQLScriptSource',
           'SQLQuerySource',
           'GenericSource', 'FileSource', 'NotebookSource',
           'Source', 'EmptySource']



import pdb
import tempfile
import subprocess
from pathlib import Path

try:
    import papermill as pm
except ImportError:
    pm = None

try:
    import jupytext
except ImportError:
    jupytext = None

try:
    import nbformat
except ImportError:
    nbformat = None

try:
    import nbconvert
except ImportError:
    nbconvert = None


from ploomber.exceptions import TaskBuildError
from ploomber.sources import NotebookSource
from ploomber.sources.NotebookSource import _cleanup_rendered_nb
from ploomber.products import File, MetaProduct
from ploomber.tasks.Task import Task
from ploomber.util import requires


def _get_exporter(nbconvert_exporter_name, extension):
    if nbconvert_exporter_name is not None:
        exporter = nbconvert.get_exporter(nbconvert_exporter_name)
    else:
        if extension == '.ipynb':
            return None
        else:
            try:
                exporter = nbconvert.get_exporter(extension.replace('.', ''))
            except ValueError:
                raise ValueError('Could not determine nbconvert exporter '
                                 'either specify in the path extension '
                                 'or pass a valid exporter name in '
                                 'the NotebookRunner constructor, '
                                 'valid exporters are: {}'
                                 .format(nbconvert.get_export_names()))

    return exporter


def _from_ipynb(path_to_nb, exporter, nbconvert_export_kwargs):

    path = Path(path_to_nb)

    nb = nbformat.reads(path.read_text(), as_version=nbformat.NO_CONVERT)
    content, _ = nbconvert.export(exporter, nb,
                                  **nbconvert_export_kwargs)

    path.write_text(content)

    return content


class NotebookRunner(Task):
    """
    Run a Jupyter notebook using papermill. Support several input formats
    via jupytext and several output formats via nbconvert

    Parameters
    ----------
    source: str or pathlib.Path
        Notebook source, if str, the content is interpreted as the actual
        notebook, if pathlib.Path, the content of the file is loaded. When
        loading from a str, ext_in must be passed
    product: ploomber.File
        The output file
    dag: ploomber.DAG
        A DAG to add this task to
    name: str, optional
        A str to indentify this task. Should not already exist in the dag
    params: dict, optional
        Notebook parameters. This are passed as the "parameters" argument
        to the papermill.execute_notebook function, by default, "product"
        and "upstream" are included
    papermill_params : dict, optional
        Other parameters passed to papermill.execute_notebook, defaults to None
    kernelspec_name: str, optional
        Kernelspec name to use, if the notebook already includes kernelspec
        data (in metadata.kernelspec), this is ignored, otherwise, the kernel
        is looked up using the jupyter_client.kernelspec.get_kernel_spec
    nbconvert_exporter_name: str, optional
        Once the notebook is run, this parameter controls whether to export
        the notebook to a different parameter using the nbconvert package,
        it is not needed unless the extension cannot be used to infer the
        final output format, in which case the nbconvert.get_exporter is used
    ext_in: str, optional
        The input extension format. If source is a pathlib.Path, the extension
        from there is used, if loaded from a str, this parameter is needed
    nb_product_key: str, optional
        If the notebook is expected to generate other products, pass the key
        to identify the output notebook (i.e. if product is a list with 3
        ploomber.File, pass the index pointing to the notebook path). If the
        only output is the notebook itself, this parameter is not needed
    static_analysis : bool
        Run static analysis after rendering.
        This requires a cell with the tag 'parameters' to exist in the
        notebook, such cell should have at least a "product = None" variable
        declared. Passed and declared parameters are compared (they make
        notebooks behave more like "functions"), pyflakes is also run to
        detect errors before executing the notebook. If the task has
        upstream dependencies an upstream parameter should also be declared
        "upstream = None"
    nbconvert_export_kwargs : dict
        Keyword arguments to pass to the nbconvert.export function (this is
        only used if exporting the output ipynb notebook to another format).
        You can use this for example to hide code cells using the exclude_input
        parameter
    local_execution : bool, optional
        Change working directory to be the parent of the notebook's source.
        Defaults to False. This resembles the default behavior when
        running notebooks interactively via `jupyter notebook`
    """
    PRODUCT_CLASSES_ALLOWED = (File, )

    @requires(['jupyter', 'papermill', 'jupytext'], 'NotebookRunner')
    def __init__(self, source, product, dag, name=None, params=None,
                 papermill_params=None, kernelspec_name='python3',
                 nbconvert_exporter_name=None, ext_in=None,
                 nb_product_key='nb', static_analysis=False,
                 nbconvert_export_kwargs=None, local_execution=False):
        self.papermill_params = papermill_params or {}
        self.nbconvert_export_kwargs = nbconvert_export_kwargs or {}
        self.kernelspec_name = kernelspec_name
        self.nbconvert_exporter_name = nbconvert_exporter_name
        self.ext_in = ext_in
        self.nb_product_key = nb_product_key
        self.static_analysis = static_analysis
        self.local_execution = local_execution
        super().__init__(source, product, dag, name, params)

        if 'cwd' in self.papermill_params and self.local_execution:
            raise KeyError('If local_execution is set to True, "cwd" should '
                           'not appear in papermill_params, as such '
                           'parameter will be set by the task itself')

        if isinstance(self.product, MetaProduct):
            if self.product.get(nb_product_key) is None:
                raise KeyError('Key "{}" does not exist in product: {}. '
                               'nb_product_key should be an existing '
                               'key to know where to save the output '
                               'notebook'.format(nb_product_key,
                                                 str(self.product)))

        if isinstance(self.product, MetaProduct):
            ext_out = self.product[self.nb_product_key].suffix
        else:
            ext_out = self.product.suffix

        self._exporter = _get_exporter(nbconvert_exporter_name, ext_out)

    def _init_source(self, source, kwargs):
        return NotebookSource(source,
                              ext_in=self.ext_in,
                              kernelspec_name=self.kernelspec_name,
                              static_analysis=self.static_analysis,
                              **kwargs)

    def develop(self):
        """
        Opens the rendered notebook (with injected parameters) and adds a
        "debugging-settings" cell to the that changes directory to the current
        active directory. This will reflect conditions when callign
        `DAG.build()`. This modified notebook is saved in the same location as
        the source with a "-tmp" added to the filename. Changes to this
        notebook can be exported to the original notebook after the notebook
        process is shut down. The "injected-parameters" and
        "debugging-settings" cells are deleted before saving.

        Notes
        -----
        If you modify the source code and call develop again, the source
        code will be updated only if the `hot_reload option` is turned on.
        See :class:`ploomber.DAGConfigurator` for details.
        """
        if self.source.loc is None:
            raise ValueError('Can only use develop in notebooks loaded '
                             'from files, not from str')

        nb = _read_rendered_notebook(self.source.rendered_nb_str)

        name = self.source.loc.name
        suffix = self.source.loc.suffix
        name_new = name.replace(suffix, '-tmp.ipynb')
        tmp = self.source.loc.with_name(name_new)
        content = nbformat.writes(nb, version=nbformat.NO_CONVERT)
        tmp.write_text(content)

        # open notebook with injected debugging cell
        _open_jupyter_notebook(str(tmp))

        # read tmp file again, to see if the user made any changes
        content_new = Path(tmp).read_text()

        # maybe exclude changes in tmp cells?
        if content == content_new:
            print('No changes found...')
        else:
            # save changes
            if _save():
                nb = nbformat.reads(content_new,
                                    as_version=nbformat.NO_CONVERT)

                # remove injected-parameters and debugging-settings cells if
                # they exist
                _cleanup_rendered_nb(nb)

                # write back in the same format and original location
                ext_source = Path(self.source.loc).suffix[1:]
                print('Saving notebook to: ', self.source.loc)
                jupytext.write(nb, self.source.loc, fmt=ext_source)
            else:
                print('Not saving changes...')

        # remove tmp file
        Path(tmp).unlink()

    def debug(self, kind='ipdb'):
        """
        Opens the notebook (with injected parameters) in debug mode in a
        temporary location
        """
        opts = {'ipdb', 'pdb'}

        if kind not in opts:
            raise ValueError('kind must be one of {}'.format(opts))

        nb = _read_rendered_notebook(self.source.rendered_nb_str)
        _, tmp_path = tempfile.mkstemp(suffix='.ipynb')
        code = jupytext.writes(nb, version=nbformat.NO_CONVERT, fmt='py')
        Path(tmp_path).write_text(code)

        if kind == 'ipdb':
            from IPython.terminal.debugger import TerminalPdb, Pdb
            code = compile(source=code, filename=tmp_path, mode='exec')

            try:
                # this seems to only work in a Terminal
                debugger = TerminalPdb()
            except AttributeError:
                # this works in a Jupyter notebook
                debugger = Pdb()
        elif kind == 'pdb':
            debugger = pdb

        try:
            debugger.run(code)
        finally:
            Path(tmp_path).unlink()

    def run(self):
        if isinstance(self.product, MetaProduct):
            path_to_out = Path(str(self.product[self.nb_product_key]))
        else:
            path_to_out = Path(str(self.product))

        # we will run the notebook with this extension, regardless of the
        # user's choice, if any error happens, this will allow them to debug
        # we will change the extension after the notebook runs successfully
        path_to_out_nb = path_to_out.with_suffix('.ipynb')

        _, tmp = tempfile.mkstemp('.ipynb')
        tmp = Path(tmp)
        tmp.write_text(self.source.rendered_nb_str)

        if self.local_execution:
            self.papermill_params['cwd'] = str(self.source.loc.parent)

        try:
            # no need to pass parameters, they are already there
            pm.execute_notebook(str(tmp), str(path_to_out_nb),
                                **self.papermill_params)
        except Exception as e:
            raise TaskBuildError('An error ocurred when calling'
                                 ' papermil.execute_notebook, partially'
                                 ' executed notebook with traceback '
                                 'available at {}'
                                 .format(str(path_to_out_nb))) from e
        finally:
            tmp.unlink()

        # if output format other than ipynb, convert using nbconvert
        # and overwrite
        if self._exporter is not None:
            path_to_out_nb.rename(path_to_out)
            _from_ipynb(path_to_out, self._exporter,
                        self.nbconvert_export_kwargs)


def _read_rendered_notebook(nb_str):
    """
    Read rendered notebook and inject cell with debugging settings
    """
    # add debug cells
    nb = nbformat.reads(nb_str, as_version=nbformat.NO_CONVERT)
    nbformat_v = nbformat.versions[nb.nbformat]

    source = """
# Debugging settings (this cell will be removed before saving)
# change the current working directory to the one when .debug() happen
# to make relative paths work
from os import chdir
chdir("{}")
""".format(Path('.').resolve())

    cell = nbformat_v.new_code_cell(source,
                                    metadata={'tags':
                                              ['debugging-settings']})
    nb.cells.insert(0, cell)

    return nb


def _open_jupyter_notebook(path):
    try:
        subprocess.call(['jupyter', 'notebook', path])
    except KeyboardInterrupt:
        print('Jupyter notebook server closed...')


def _save():
    res = input('Notebook changed, do you want to save changes '
                'in the original location? (injected parameters '
                'and debugging cells will be removed before '
                'saving). Enter "no" to skip saving changes, '
                'anything else will be interpreted as "yes": ')
    return res != 'no'



"""
A mapping object with text and HTML representations
"""
from copy import deepcopy
from pathlib import Path
import tempfile
from collections.abc import Mapping, Iterable
from tabulate import tabulate


def _is_iterable(obj):
    return isinstance(obj, Iterable) and not isinstance(obj, str)


class Row:
    """A class to represent a dictionary as a table row
    """

    def __init__(self, mapping):
        if not isinstance(mapping, Mapping):
            raise TypeError('Rows must be initialized with mappings')

        self._mapping = mapping
        self._str = tabulate([self._mapping], headers='keys',
                             tablefmt='simple')
        self._html = tabulate([self._mapping], headers='keys',
                              tablefmt='html')

    def __str__(self):
        return self._str

    def __repr__(self):
        return str(self)

    def _repr_html_(self):
        return self._html

    def __getitem__(self, key):
        if _is_iterable(key):
            return Row({k: self._mapping[k] for k in key})
        else:
            return self._mapping[key]

    def __setitem__(self, key, value):
        self._mapping[key] = value

    def __eq__(self, other):
        return self._mapping == other

    @property
    def columns(self):
        return tuple(self._mapping.keys())


class Table:
    """A collection of rows
    """

    def __init__(self, rows):
        # TODO: remove this, only use ._values
        self._rows = self.data_preprocessing(rows)
        self._values = self._transform(rows)
        self._str = tabulate(self.values, headers='keys',
                             tablefmt='simple')
        self._html = tabulate(self.values, headers='keys',
                              tablefmt='html')

    def _transform(self, rows):
        """Convert [{key: value}, {key: value2}] to [{key: [value, value2]}]
        """
        if not len(rows):
            return {}

        cols_combinations = set(tuple(sorted(row.columns)) for row in rows)

        if len(cols_combinations) > 1:
            raise KeyError('All rows should have the same columns, got: '
                           '{}'.format(cols_combinations))

        columns = rows[0].columns

        return {col: [row[col] for row in rows] for col in columns}

    def __str__(self):
        return self._str

    def __repr__(self):
        return str(self)

    def _repr_html_(self):
        return self._html

    def __getitem__(self, key):
        if _is_iterable(key):
            return Table([row[key] for row in self._rows])
        else:
            return self.values[key]

    def __iter__(self):
        for col in self.values:
            yield col

    def __len__(self):
        return len(self._values.keys())

    def __eq__(self, other):
        return self._rows == other

    def data_preprocessing(self, data):
        return data

    def save(self, path=None):
        if path is None:
            path = Path(tempfile.mktemp(suffix='.html'))

        path.write_text(self._html)

        return path

    def to_format(self, fmt):
        return tabulate(self.values, headers='keys', tablefmt=fmt)

    def to_pandas(self):
        import pandas as pd
        return pd.DataFrame(self.values)

    def to_dict(self):
        return deepcopy(self.values)

    @property
    def values(self):
        return self._values


class TaskReport(Row):

    @classmethod
    def with_data(cls, name, ran, elapsed):
        return cls({'name': name,
                    'Ran?': ran,
                    'Elapsed (s)': elapsed})

    @classmethod
    def empty_with_name(cls, name):
        return cls.with_data(name, False, 0)


class BuildReport(Table):
    """A Table that adds a columns for checking task elapsed time
    """

    def data_preprocessing(self, rows):
        """Create a build report from several tasks
        """
        total = sum([row['Elapsed (s)'] or 0 for row in rows])

        def compute_pct(elapsed, total):
            if not elapsed:
                return 0
            else:
                return 100 * elapsed / total

        for row in rows:
            row['Percentage'] = compute_pct(row['Elapsed (s)'], total)

        return rows



def keys(valid, passed, required=None, name='spec'):
    passed = set(passed)

    if valid:
        extra = passed - set(valid)

        if extra:
            raise KeyError("Error validating {}, the following keys aren't "
                           "valid: {}. Valid keys are: {}"
                           .format(name, extra, valid))

    if required:
        missing = set(required) - passed

        if missing:
            raise KeyError("Error validating {}, the following required "
                           "keys are missing: {}".format(name, missing))



from ploomber.clients.Client import Client
from ploomber.clients.db import (DBAPIClient, SQLAlchemyClient,
                                 DrillClient)
from ploomber.clients.shell import ShellClient, RemoteShellClient

__all__ = ['Client', 'DBAPIClient', 'SQLAlchemyClient', 'DrillClient',
           'ShellClient', 'RemoteShellClient']



"""
API:

on_finish:
  data:
    column:
      - unique
      - no_nulls
      - values: [a, b, c]
"""


class TestSpec:
    """Convert dictionaries to a callable that runs assertions on products
    """
    def __init__(self, data, arg, client=None):
        pass

    def __call__(self):
        pass



from copy import deepcopy
from pathlib import Path

from jinja2 import Template

from ploomber.tasks import PythonCallable
from ploomber.tasks.tasks import _Gather, _Partition
from ploomber.products import File
from ploomber.tasks.TaskGroup import TaskGroup


class PartitionedFile(File):
    @property
    def _path_to_stored_source_code(self):
        # point to the parent folder
        return Path(str(self._path_to_file.parent) + '.source')

    def save_metadata(self):
        # do not save metadata
        pass

    # I think i have to remove this
    def _outdated_code_dependency(self):
        return False


def partitioned_execution(upstream_partitioned,
                          downstream_callable,
                          downstream_prefix,
                          downstream_path,
                          partition_ids,
                          partition_template='partition={{id}}',
                          upstream_other=None):
    # make sure output is a File
    assert isinstance(upstream_partitioned.product, File)

    dag = upstream_partitioned.dag
    partition_template_w_suffix = Template(
        downstream_prefix + '_' + partition_template)
    partition_template_t = Template(partition_template)

    # instantiate null tasks
    nulls = [_Partition
             (product=PartitionedFile(Path(str(upstream_partitioned.product),
                                           partition_template_t.render(id=id_))),
              dag=dag,
              name=Template('null_'+partition_template).render(id=id_))
             for id_ in partition_ids]

    def make_file(id_):
        f = PartitionedFile(Path(downstream_path,
                                 partition_template_t.render(id=id_)))
        return f

    # TODO: validate downstream product is File
    tasks = [PythonCallable(downstream_callable,
                            product=make_file(id_),
                            dag=dag,
                            name=(partition_template_w_suffix
                                  .render(id=id_)),
                            params={'upstream_key':
                                    (Template('null_'+partition_template)
                                     .render(id=id_))}
                            )
             for id_ in partition_ids]

    gather = _Gather(source=downstream_callable,
                     product=File(downstream_path),
                     dag=dag,
                     name=downstream_prefix)

    # TODO: "fuse" operator that merges task chains and shows it like a single
    # task - maybe this should be like this and the dag.plot function take
    # care of simplifications for viz purposes?
    for null, task in zip(nulls, tasks):
        upstream_partitioned >> null >> task >> gather

        if upstream_other:
            upstream_other >> task

    return gather


def make_task_group(task_class, task_kwargs, dag, name, params_array,
                    namer=None):
    # validate task_kwargs
    if 'dag' in task_kwargs:
        raise KeyError('dag should not be part of task_kwargs')

    if 'name' in task_kwargs:
        raise KeyError('name should not be part of task_kwargs')

    if 'params' in task_kwargs:
        raise KeyError('params should not be part of task_kwargs')

    if 'product' not in task_kwargs:
        raise KeyError('product should be in task_kwargs')

    # TODO: validate {{index}} appears in product - maybe all products
    # should have a way to extract which placeholders exist?

    tasks_all = []

    for i, params in enumerate(params_array):

        # each task should get a different deep copy, primarily cause they
        # should have a different product
        kwargs = deepcopy(task_kwargs)
        # params should also be different copies, otherwise if the same
        # grid is re-used in several tasks, modifying anything there will
        # have side-effects
        params = deepcopy(params)

        if namer:
            task_name = namer(params)
            params['name'] = task_name
        else:
            task_name = name+str(i)
            params['name'] = i

        t = task_class(**kwargs,
                       dag=dag,
                       name=task_name,
                       params=params)
        tasks_all.append(t)

    return TaskGroup(tasks_all, False, name)



from copy import copy
from collections import abc


class Params(abc.Mapping):
    """
    Read-only mapping to represent params passed in Task constructor. It
    initializes with a copy of the passed dictionary. Cannot be initialized
    with a key "upstream" nor "product" as they are added upon Task rendering
    """

    def __init__(self, params=None):
        if params is None:
            self._dict = {}
        else:
            if 'upstream' in params:
                raise ValueError('Task params cannot be initialized with an '
                                 '"upstream" key as it automatically added '
                                 'upon rendering')

            if 'product' in params:
                raise ValueError('Task params cannot be initialized with an '
                                 '"product" key as it automatically added '
                                 'upon rendering')

            self._dict = copy(params)

    def to_dict(self):
        return copy(self._dict)

    def __getitem__(self, key):
        try:
            return self._dict[key]
        except KeyError:
            raise KeyError('Cannot obtain Task param named '
                           '"{}", declared params are: {}'
                           .format(key, list(self._dict.keys())))

    def __setitem__(self, key, value):
        raise RuntimeError('Task params are read-only, if you need a copy'
                           ' use Params.to_dict() (returns a shallow copy)'
                           ' of the underlying dictionary')

    def __iter__(self):
        for name in self._dict.keys():
            yield name

    def __len__(self):
        return len(self._dict)

    def __str__(self):
        return str(self._dict)

    def __repr__(self):
        return 'Params({})'.format(repr(self._dict))

    def get(self, key):
        return self._dict.get(key)

    def __delitem__(self, key):
        del self._dict[key]



"""
Get the raw data and save it

Data reference: https://archive.ics.uci.edu/ml/datasets/Adult
"""
import pandas as pd

# + tags=["parameters"]
upstream = None
product = {'nb': 'output/data.ipynb', 'data': 'output/data.csv'}
# -

# +
url = ('https://archive.ics.uci.edu/ml/machine-learning-databases/adult'
       '/adult.data')
df = pd.read_csv(url, header=None)
df.columns = [
    'age', 'workclass', 'fnlwgt', 'education', 'education-num',
    'marital-status', 'occupation', 'relationship', 'race', 'sex',
    'capital-gain', 'capital-loss', 'hours-per-week', 'native-country',
    'income'
]
df.head()
# -

# +
df.to_csv(str(product['data']), index=False)
# -



"""
Utils for comparing source code
"""
import io
import tokenize
import warnings
from difflib import Differ
import parso

try:
    import sqlparse
except ImportError:
    sqlparse = None


try:
    import autopep8
except ImportError:
    autopep8 = None


def normalize_null(code):
    return code


def normalize_sql(code):
    if not sqlparse:
        raise ImportError('sqlparse is required for normalizing SQL code')

    return None if code is None else sqlparse.format(code,
                                                     keyword_case='upper',
                                                     identifier_case='lower',
                                                     strip_comments=True,
                                                     reindent=True,
                                                     indent_with=4)


def _delete_python_comments(code):
    tokens = tokenize.generate_tokens(io.StringIO(code).readline)
    tokens = [(num, val) for num, val, _, _, _ in tokens
              if num != tokenize.COMMENT]
    return tokenize.untokenize(tokens)


def normalize_python(code):
    # TODO: we should really be comparing the tree between a, b but this
    # works for now

    if code is None:
        return None

    code = _delete_python_comments(code)

    if not autopep8 or not parso:
        raise ImportError('autopep8 and parso are required for normalizing '
                          'Python code: pip install autopep8 parso')

    try:
        doc_node = parso.parse(code).children[0].get_doc_node()
    except Exception as e:
        warnings.warn('Could not remove docstring from Python code: {}'
                      .format(e))
    else:
        if doc_node is not None:
            code = code.replace(doc_node.get_code(), '')

    code = autopep8.fix_code(code)

    return code


def diff_strings(a, b):
    """Compute the diff between two strings
    """
    d = Differ()

    if a is None and b is None:
        return '[Both a and b are None]'

    out = ''

    if a is None:
        out += '[a is None]\n'
    elif b is None:
        out += '[a is None]\n'

    a = '' if a is None else a
    b = '' if b is None else b

    result = d.compare(a.splitlines(keepends=True),
                       b.splitlines(keepends=True))
    out += ''.join(result)

    return out


class CodeDiffer:
    NORMALIZERS = {None: normalize_null, 'py': normalize_python,
                   'sql': normalize_sql}

    def is_different(self, a, b, extension=None):
        normalizer = self._get_normalizer(extension)

        a_norm = normalizer(a)
        b_norm = normalizer(b)

        diff = self.get_diff(a_norm, b_norm, normalize=False)

        return a_norm != b_norm, diff

    def get_diff(self, a, b, extension=None, normalize=True):
        if normalize:
            normalizer = self._get_normalizer(extension)

            a = normalizer(a)
            b = normalizer(b)

        diff = diff_strings(a, b)

        if extension is not None:
            diff = '[Code was normalized]\n' + diff

        return diff

    def _get_normalizer(self, extension):
        if extension in self.NORMALIZERS:
            return self.NORMALIZERS[extension]
        else:
            return normalize_null



"""
Clients that communicate with databases
"""
try:
    from sqlalchemy import create_engine
except ImportError:
    create_engine = None

from ploomber.clients.Client import Client
from ploomber.util import requires

import re
from urllib.parse import urlparse, urlunparse


def code_split(code, token=';'):
    only_whitespace = re.compile(r'^\s*$')

    for part in code.split(token):
        if not re.match(only_whitespace, part):
            yield part


def safe_uri(parsed):
    password = parsed.password

    if password:
        parts = list(parsed)
        # netloc
        parts[1] = parts[1].replace(password, '********')
        return urlunparse(parts)
    else:
        return urlunparse(parsed)


class DBAPIClient(Client):
    """A client for a PEP 249 compliant client library

    Parameters
    ----------
    connect_fn : callable
        The function to use to open the connection

    connect_kwargs : dict
        Keyword arguments to pass to connect_fn

    split_source : str, optional
        Some database drivers do not support multiple commands, use this
        optiion to split commands by a given character (e.g. ';') and send
        them one at a time. Defaults to False (no spltting)
    """
    def __init__(self, connect_fn, connect_kwargs, split_source=False):
        super().__init__()
        self.connect_fn = connect_fn
        self.connect_kwargs = connect_kwargs
        self.split_source = split_source

        # there is no open connection by default
        self._connection = None

    @property
    def connection(self):
        """Return a connection, open one if there isn't any
        """
        # if there isn't an open connection, open one...
        if self._connection is None:
            self._connection = self.connect_fn(**self.connect_kwargs)

        return self._connection

    def execute(self, code):
        """Execute code with the existing connection
        """
        cur = self.connection.cursor()

        if self.split_source:
            for command in code_split(code, token=self.split_source):
                cur.execute(command)
        else:
            cur.execute(code)

        self.connection.commit()
        cur.close()

    def close(self):
        """Close connection if there is an active one
        """
        if self._connection is not None:
            self._connection.close()

    def __getstate__(self):
        state = super().__getstate__()
        state['_connection'] = None
        return state


class SQLAlchemyClient(Client):
    """Client for connecting with any SQLAlchemy supported database

    Parameters
    ----------
    uri: str
        URI to pass to sqlalchemy.create_engine

    Notes
    -----
    SQLite client does not support sending more than one command at a time,
    if using such backend code will be split and several calls to the db
    will be performed.
    """
    split_source = ['sqlite']

    @requires(['sqlalchemy'], 'SQLAlchemyClient')
    def __init__(self, uri):
        super().__init__()
        self._uri = uri
        self._uri_parsed = urlparse(uri)
        self._uri_safe = safe_uri(self._uri_parsed)
        self.flavor = self._uri_parsed.scheme
        self._engine = None
        self._connection = None

    @property
    def connection(self):
        """Return a connection from the pool
        """
        # we have to keep this reference here,
        # if we just return self.engine.raw_connection(),
        # any cursor from that connection will fail
        # doing: engine.raw_connection().cursor().execute('') fails!
        if self._connection is None:
            self._connection = self.engine.raw_connection()

        # if a task or product calls client.connection.close(), we have to
        # re-open the connection
        if not self._connection.is_valid:
            self._connection = self.engine.raw_connection()

        return self._connection

    def execute(self, code):
        cur = self.connection.cursor()

        if self.flavor in self.split_source:
            for command in code_split(code):
                cur.execute(command)
        else:
            cur.execute(code)

        self.connection.commit()
        cur.close()

    def close(self):
        """Closes all connections
        """
        self._logger.info('Disposing engine %s', self._engine)
        if self._engine is not None:
            self._engine.dispose()
            self._engine = None

    @property
    def engine(self):
        """Returns a SQLAlchemy engine
        """
        if self._engine is None:
            self._engine = create_engine(self._uri)

        return self._engine

    def __getstate__(self):
        state = super().__getstate__()
        state['_engine'] = None
        state['_connection'] = None
        return state

    def __str__(self):
        return self._uri_safe

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, self._uri_safe)


class DrillClient(Client):
    def __init__(self, params=dict(host='localhost', port=8047)):
        self.params = params
        self._set_logger()
        self._connection = None

    @property
    def connection(self):
        from pydrill.client import PyDrill

        if self._connection is None:
            self._connection = PyDrill(**self.params)

        return self._connection

    def execute(self, code):
        """Run code
        """
        return self.connection.query(code)

    def close(self):
        pass



"""
A Product specifies a persistent object in disk such as a file in the local
filesystem or an table in a database. Each Product is uniquely identified,
for example a file can be specified using a absolute path, a table can be
fully specified by specifying a database, a schema and a name. Names
are lazy evaluated, they can be built from templates
"""
import abc
import logging

from ploomber.products.Metadata import Metadata


def _prepare_metadata(metadata):
    return metadata


class Product(abc.ABC):
    """
    Abstract class for all Products

    Attributes
    ----------
    prepare_metadata : callable
        A hook to execute before saving metadata, should include a "metadata"
        parameter and might include "product". "metadata" will be a
        dictionary with the metadata to save, it is not recommended to change
        any of the existing keys but additional key-value pairs might be
        included
    """

    def __init__(self, identifier):
        self._identifier = self._init_identifier(identifier)

        if self._identifier is None:
            raise TypeError('_init_identifier must return a value, returned '
                            'None')

        self.task = None
        self.logger = logging.getLogger('{}.{}'.format(__name__,
                                                       type(self).__name__))

        self._outdated_data_dependencies_status = None
        self._outdated_code_dependency_status = None
        self.metadata = Metadata(self)

        self.prepare_metadata = _prepare_metadata

    def _save_metadata(self, source_code):
        # this is called by task in the exec_status setter
        self.metadata.update(source_code)

    @property
    def task(self):
        if self._task is None:
            raise ValueError('This product has not been assigned to any Task')

        return self._task

    @task.setter
    def task(self, value):
        self._task = value

    def render(self, params, **kwargs):
        """
        Render Product - this will render contents of Templates used as
        identifier for this Product, if a regular string was passed, this
        method has no effect
        """
        self._identifier.render(params, **kwargs)

    def _is_outdated(self, outdated_by_code=True):
        """
        Given current conditions, determine if the Task that holds this
        Product should be executed

        Returns
        -------
        bool
            True if the Task should execute
        """
        run = False

        self.logger.info('Checking status for task "%s"', self.task.name)

        # check product...
        p_exists = self.exists()

        # check dependencies only if the product exists
        if p_exists:

            if self._outdated_data_dependencies():
                run = True
                self.logger.info('Outdated data deps...')
            else:
                self.logger.info('Up-to-date data deps...')

            if outdated_by_code:
                if self._outdated_code_dependency():
                    run = True
                    self.logger.info('Outdated code dep...')
                else:
                    self.logger.info('Up-to-date code dep...')
        else:
            run = True

            # just log why it will run
            if not p_exists:
                self.logger.info('Product does not exist...')

        self.logger.info('Should run? %s', run)

        return run

    def _outdated_data_dependencies(self):
        """
        Determine if the product is outdated by checking upstream timestamps
        """

        if self._outdated_data_dependencies_status is not None:
            self.logger.debug(('Returning cached data dependencies status. '
                               'Outdated? %s'),
                              self._outdated_data_dependencies_status)
            return self._outdated_data_dependencies_status

        def is_outdated(up_prod):
            """
            A task becomes data outdated if an upstream product has a higher
            timestamp or if an upstream product is outdated
            """
            if (self.metadata.timestamp is None
                    or up_prod.metadata.timestamp is None):
                return True
            else:
                return ((up_prod.metadata.timestamp > self.metadata.timestamp)
                        or up_prod._is_outdated())

        outdated = any([is_outdated(up.product) for up
                        in self.task.upstream.values()])

        self._outdated_data_dependencies_status = outdated

        self.logger.debug(('Finished checking data dependencies status. '
                           'Outdated? %s'),
                          self._outdated_data_dependencies_status)

        return self._outdated_data_dependencies_status

    def _outdated_code_dependency(self):
        """
        Determine if the product is outdated by checking the source code that
        it generated it
        """
        if self._outdated_code_dependency_status is not None:
            self.logger.debug(('Returning cached code dependencies status. '
                               'Outdated? %s'),
                              self._outdated_code_dependency_status)
            return self._outdated_code_dependency_status

        outdated, diff = self.task.dag.differ.is_different(
            self.metadata.stored_source_code,
            str(self.task.source),
            extension=self.task.source.extension)

        self._outdated_code_dependency_status = outdated

        self.logger.debug(('Finished checking code status for task "%s" '
                           'Outdated? %s'),
                          self.task.name,
                          self._outdated_code_dependency_status)

        if outdated:
            self.logger.info('Task "%s" has outdated code. Diff:\n%s',
                             self.task.name, diff)

        return self._outdated_code_dependency_status

    def _clear_cached_status(self):
        # These flags keep a cache of the Product's outdated status, they
        # are computed using the Product's metadata, hence they will only
        # change when the metadata changes. Metadata changes in three
        # situations: 1) at startup (loaded from disk), 2) after build
        # (metadata is updated and then saved to disk) and 3) Forced load
        # (if we force loading, currently not implemented).
        self._outdated_data_dependencies_status = None
        self._outdated_code_dependency_status = None

    def __str__(self):
        return str(self._identifier)

    def __repr__(self):
        return '{}({})'.format(type(self).__name__,
                               self._identifier.best_str(shorten=True))

    def __getstate__(self):
        state = self.__dict__.copy()
        # logger is not pickable, so we remove them and build
        # them again in __setstate__
        del state['logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = logging.getLogger('{}.{}'.format(__name__,
                                                       type(self).__name__))

    def to_json_serializable(self):
        """Returns a JSON serializable version of this product
        """
        # NOTE: this is used in tasks where only JSON serializable parameters
        # are supported such as NotebookRunner that depends on papermill
        return str(self)

    def __len__(self):
        # MetaProduct return the number of products, this is a single Product
        # hence the 1
        return 1

    # Subclasses must implement the following methods

    @abc.abstractmethod
    def _init_identifier(self, identifier):
        pass

    @abc.abstractmethod
    def fetch_metadata(self):
        pass

    @abc.abstractmethod
    def save_metadata(self, metadata):
        pass

    @abc.abstractmethod
    def exists(self):
        """
        This method returns True if the product exists, it is not part
        of the metadata, so there is no cached status
        """
        pass

    @abc.abstractmethod
    def delete(self, force=False):
        """Deletes the product
        """
        pass



"""
Generate plot
"""
import pandas as pd

# + tags=["parameters"]
upstream = ['clean']
product = {'nb': 'output/plot.ipynb'}
# -

# +
df = pd.read_csv(upstream['clean']['data'])
grouped = df.groupby('sex')[['age', 'hours-per-week']].mean()
grouped.columns = ['Mean age', 'Mean hours per week worked']
grouped.head()
# -

ax = grouped.plot.barh()
# -



from logging import NullHandler
import logging
from ploomber.dag.DAG import DAG
from ploomber.dag.DAGConfigurator import DAGConfigurator
from ploomber.env.env import Env
from ploomber.env.decorators import load_env, with_env
from ploomber.placeholders.SourceLoader import SourceLoader
from ploomber.jupyter import _load_jupyter_server_extension


__version__ = '0.6.1'

# Set default logging handler to avoid "No handler found" warnings.

logging.getLogger(__name__).addHandler(NullHandler())


__all__ = ['DAG', 'Env', 'SourceLoader', 'load_env', 'with_env',
           'DAGConfigurator']


def _jupyter_server_extension_paths():
    return [{'module': 'ploomber'}]


load_jupyter_server_extension = _load_jupyter_server_extension



from ploomber.util.util import (safe_remove, image_bytes2html,
                                isiterable, path2fig,
                                requires)
from ploomber.util import markup
from ploomber.util.param_grid import Interval, ParamGrid

__all__ = ['safe_remove', 'image_bytes2html', 'Interval',
           'ParamGrid', 'isiterable', 'path2fig', 'requires', 'markup']



import re
import sys
from pathlib import Path

from jinja2 import Environment, PackageLoader
import click
from ploomber import __version__
from ploomber.entry import entry as entry_module
from ploomber.spec.DAGSpec import DAGSpec

try:
    import importlib.resources as resources
except ImportError:
    import importlib_resources as resources


def _is_valid_name(package_name):
    match = re.match(r'^[\w-]+$', package_name) or False
    return match and not package_name[0].isnumeric()


@click.group()
@click.version_option(version=__version__)
def cli():
    """Ploomber command line interface.

    To start an nteractive session (use "dag" variable when it starts):

    ipython -i -m ploomber.entry pipeline.yaml -- --action status
    """
    pass


@cli.command()
def new():
    """Create a new project
    """
    _new()


@cli.command()
def add():
    """Add tasks declared in pipeline.yaml whose source code files do not exist
    """
    _add()


class FileLoader:
    def __init__(self, directory, project_name=None):
        self.env = Environment(loader=PackageLoader(
            'ploomber', str(Path('resources', directory))),
                               variable_start_string='[[',
                               variable_end_string=']]',
                               block_start_string='[%',
                               block_end_string='%]')
        self.directory = directory
        self.project_name = project_name

    def get_template(self, name):
        return self.env.get_template(name)

    def copy(self, name):
        module = '.'.join(['ploomber', 'resources', self.directory])
        content = resources.read_text(module, name)
        Path(self.project_name, name).write_text(content)


def _add():
    spec, path = DAGSpec.auto_load(to_dag=False)
    env = FileLoader('ploomber_add')

    if path:
        click.echo('Found spec at {}'.format(path))

        for task in spec['tasks']:
            source = Path(task['source'])

            if not source.exists():
                # create parent folders if needed
                source.parent.mkdir(parents=True, exist_ok=True)

                if source.suffix in {'.py', '.sql'}:
                    click.echo('Adding {}...'.format(source))
                    template = env.get_template('task' + source.suffix)
                    content = template.render(**spec['meta'])
                    source.write_text(content)

                else:
                    click.echo('Error: This command does not support adding '
                               'tasks with extension "{}", valid ones are '
                               '.py and .sql. Skipping {}'.format(
                                   source.suffix, source))

    else:
        click.echo('Error: No pipeline.yaml spec found...')


def _new():
    click.echo('This utility will guide you through the process of creating '
               'a new project')

    valid_name = False

    while not valid_name:
        name = click.prompt(
            'Enter a name for your project (allowed: '
            'alphanumeric, underscores and hyphens)',
            type=str)
        valid_name = _is_valid_name(name)

        if not valid_name:
            click.echo('"%s" is not a valid project name, choose another.' %
                       name)

    env = FileLoader('ploomber_new', project_name=name)

    click.echo('Creating %s/' % name)
    Path(name).mkdir()

    db = click.confirm('Do you need to connect to a database?')

    if db:
        click.echo('Adding db.py...')
        env.copy('db.py')

    click.echo('Adding pipeline.yaml...')
    content = env.get_template('pipeline.yaml').render(db=db)
    Path(name, 'pipeline.yaml').write_text(content)

    conda = click.confirm('Do you you want to use conda to '
                          'manage virtual environments (recommended)?')

    if conda:
        # check if conda is installed...
        click.echo('Adding environment.yml...')
        content = env.get_template('environment.yml').render(name=name)
        Path(name, 'environment.yml').write_text(content)

    click.echo('Adding README.md...')
    content = env.get_template('README.md').render(name=name,
                                                   db=db,
                                                   conda=conda)
    path_to_readme = Path(name, 'README.md')
    path_to_readme.write_text(content)

    click.echo('Adding raw.py, clean.py and plot.py...')
    env.copy('raw.py')
    env.copy('clean.py')
    env.copy('plot.py')
    click.echo('Done!')

    Path(name, 'output').mkdir()
    click.echo('Check out {} to get started'.format(path_to_readme))


@cli.command()
@click.argument('entry_point')
def entry(entry_point):
    """Call an entry point (pipeline.yaml or dotted path to factory)
    """
    # NOTE: we don't use the argument here, it is parsed by _main
    # pop the second element ('entry') to make the CLI behave as expected
    sys.argv.pop(1)
    # Add the current working directory, this is done automatically when
    # calling "python -m ploomber.entry" but not here ("ploomber entry")
    sys.path.append('')
    entry_module._main()



import json
import subprocess
from subprocess import CalledProcessError
from shlex import quote
import sys
from pathlib import Path


def _run_command(path, command):
    """Safely run command in certain path
    """
    path = str(path)

    if not Path(path).is_dir():
        raise ValueError('{} is not a directory'.format(path))

    command_ = 'cd {path} && {cmd}'.format(path=quote(path), cmd=command)

    out = subprocess.check_output(command_, shell=True)
    s = out.decode('utf-8')

    # remove trailing \n
    if s[-1:] == '\n':
        s = s[:-1]

    return s


def one_line_git_summary(path):
    """Get one line git summary"""
    return _run_command(path, 'git show --oneline -s')


def git_hash(path):
    """Get git hash"""
    return _run_command(path, 'git describe --tags --always --dirty=-dirty')


def get_git_timestamp(path):
    """Timestamp for last commit
    """
    return int(_run_command(path, 'git log -1 --format=%ct'))


def current_branch(path):
    # seems like the most reliable way is to do:
    # git branch --show-current, but that was added in a recent git
    # version 2.22, for older versions, the one below works
    try:
        return _run_command(path, 'git symbolic-ref --short HEAD')
    except CalledProcessError:
        # if detached head, the command above does not work, since there is
        # no current branch
        return None


def get_version(package_name):
    """Get package version
    """
    installation_path = sys.modules[package_name].__file__

    NON_EDITABLE = True if 'site-packages/' in installation_path else False

    if NON_EDITABLE:
        return getattr(package_name, '__version__')
    else:
        parent = str(Path(installation_path).parent)

    return one_line_git_summary(parent)


def get_diff(path):
    return _run_command(path, "git diff -- . ':(exclude)*.ipynb'")


def get_env_metadata(path):
    git_summary = one_line_git_summary(path)
    hash_ = git_hash(path)
    git_diff = get_diff(path)
    git_timestamp = get_git_timestamp(path)
    git_branch = current_branch(path)

    git_location = git_branch or hash_

    return dict(git_summary=git_summary, git_hash=hash_, git_diff=git_diff,
                git_timestamp=git_timestamp, git_branch=git_branch,
                git_location=git_location)


def save_env_metadata(env, path_to_output):
    summary = one_line_git_summary(env.path.home)
    hash_ = git_hash(env.path.home)
    diff = get_diff(env.path.home)

    metadata = dict(summary=summary, hash=hash_)
    path_to_patch_file = Path(path_to_output).with_suffix('.patch')

    with open(path_to_output, 'w') as f:
        json.dump(metadata, f)

    with open(path_to_patch_file, 'w') as f:
        f.writelines(diff)



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# Add description for your Python task here

# + tags=["parameters"]
[% if extract_upstream -%]
# extract_upstream is set to True in your pipeline.yaml. If this task has
# upstream dependencies, list their names here (e.g. {'some_task',
# 'another_task'}. If there are no dependencies, leave it as None
upstream = None
[% else -%]
# extract_upstream is set to False in your pipeline.yaml, keep it equal to
# None here and declare "upstream" dependencies directly in the YAML spec
upstream = None
[% endif %]
[% if extract_product -%]
# extract_product is set to True in your pipeline.yaml. Product must be the
# path to save the executed version of this script (a notebook), if this task
# generates other files, declare a dictionary with a key "nb" to point to the
# executed notebook and add other keys for the rest of the files
product = {'nb': 'path/to/notebook.ipynb', 'data': 'path/to/output.csv'}
[% else -%]
# extract_product is set to False in your pipeline.yaml, keep it equal to
# None here and declare a "product" directly in the YAML spec
product = None
[% endif %]
# NOTE: During execution, a cell will be injected below to resolve "upstream"
# and "product" based on preferences in your pipeline.yaml spec. If you
# activated the jupyter notebook extension, you'll see the injected cell
# as well when you open this file from the jupyter notebook app

# +
# your code here...



import importlib
import inspect

from ploomber.sources.sources import Source
from ploomber.util.util import signature_check


class CallableLoader:

    def __init__(self, callable_, hot_reload):
        self.hot_reload = hot_reload

        if self.hot_reload:
            self.module_name = inspect.getmodule(callable_).__name__
            self.name = callable_.__name__
        else:
            self.callable_ = callable_

    def __call__(self):
        if self.hot_reload:
            module = importlib.import_module(self.module_name)
            importlib.reload(module)
            return getattr(module, self.name)
        else:
            return self.callable_


class PythonCallableSource(Source):
    """
    A source object to encapsulate a Python callable (i.e. functions).
    """

    def __init__(self, primitive, hot_reload=False):
        if not callable(primitive):
            raise TypeError('{} must be initialized'
                            'with a Python callable, got '
                            '"{}"'
                            .format(type(self).__name__),
                            type(primitive).__name__)
        self._callable_loader = CallableLoader(primitive, hot_reload)
        self._source_as_str = None
        self._loc = None
        self._hot_reload = hot_reload
        self.__source_lineno = None

    @property
    def primitive(self):
        return self._callable_loader()

    @property
    def _source_lineno(self):
        if self.__source_lineno is None or self._hot_reload:
            _, self.__source_lineno = inspect.getsourcelines(self.primitive)

        return self.__source_lineno

    def __str__(self):
        if self._source_as_str is None or self._hot_reload:
            self._source_as_str = inspect.getsource(self.primitive)

        return self._source_as_str

    @property
    def doc(self):
        return self.primitive.__doc__

    @property
    def loc(self):
        if self._loc is None or self._hot_reload:
            self._loc = inspect.getsourcefile(self.primitive)

        return '{}:{}'.format(self._loc, self._source_lineno)

    def render(self, params):
        self._post_render_validation(None, params)

    @property
    def extension(self):
        return 'py'

    @property
    def name(self):
        return self.primitive.__name__

    def _post_render_validation(self, rendered_value, params):
        """
        Validation function executed after rendering
        """
        signature_check(self.primitive, params, self.name)

    def _post_init_validation(self, value):
        # TODO: verify the callable has a product parameter
        pass

    @property
    def variables(self):
        raise NotImplementedError
